import pytest
from pgcli.packages.parseutils import (
    is_destructive,
    parse_destructive_warning,
    BASE_KEYWORDS,
    ALL_KEYWORDS,
)
from pgcli.packages.parseutils.tables import extract_tables
from pgcli.packages.parseutils.utils import find_prev_keyword, is_open_quote


def test_empty_string():
    tables = extract_tables("")
    assert tables == ()


def test_simple_select_single_table():
    tables = extract_tables("select * from abc")
    assert tables == ((None, "abc", None, False),)


@pytest.mark.parametrize(
    "sql", ['select * from "abc"."def"', 'select * from abc."def"']
)
def test_simple_select_single_table_schema_qualified_quoted_table(sql):
    tables = extract_tables(sql)
    assert tables == (("abc", "def", '"def"', False),)


@pytest.mark.parametrize("sql", ["select * from abc.def", 'select * from "abc".def'])
def test_simple_select_single_table_schema_qualified(sql):
    tables = extract_tables(sql)
    assert tables == (("abc", "def", None, False),)


def test_simple_select_single_table_double_quoted():
    tables = extract_tables('select * from "Abc"')
    assert tables == ((None, "Abc", None, False),)


def test_simple_select_multiple_tables():
    tables = extract_tables("select * from abc, def")
    assert set(tables) == {(None, "abc", None, False), (None, "def", None, False)}


def test_simple_select_multiple_tables_double_quoted():
    tables = extract_tables('select * from "Abc", "Def"')
    assert set(tables) == {(None, "Abc", None, False), (None, "Def", None, False)}


def test_simple_select_single_table_deouble_quoted_aliased():
    tables = extract_tables('select * from "Abc" a')
    assert tables == ((None, "Abc", "a", False),)


def test_simple_select_multiple_tables_deouble_quoted_aliased():
    tables = extract_tables('select * from "Abc" a, "Def" d')
    assert set(tables) == {(None, "Abc", "a", False), (None, "Def", "d", False)}


def test_simple_select_multiple_tables_schema_qualified():
    tables = extract_tables("select * from abc.def, ghi.jkl")
    assert set(tables) == {("abc", "def", None, False), ("ghi", "jkl", None, False)}


def test_simple_select_with_cols_single_table():
    tables = extract_tables("select a,b from abc")
    assert tables == ((None, "abc", None, False),)


def test_simple_select_with_cols_single_table_schema_qualified():
    tables = extract_tables("select a,b from abc.def")
    assert tables == (("abc", "def", None, False),)


def test_simple_select_with_cols_multiple_tables():
    tables = extract_tables("select a,b from abc, def")
    assert set(tables) == {(None, "abc", None, False), (None, "def", None, False)}


def test_simple_select_with_cols_multiple_qualified_tables():
    tables = extract_tables("select a,b from abc.def, def.ghi")
    assert set(tables) == {("abc", "def", None, False), ("def", "ghi", None, False)}


def test_select_with_hanging_comma_single_table():
    tables = extract_tables("select a, from abc")
    assert tables == ((None, "abc", None, False),)


def test_select_with_hanging_comma_multiple_tables():
    tables = extract_tables("select a, from abc, def")
    assert set(tables) == {(None, "abc", None, False), (None, "def", None, False)}


def test_select_with_hanging_period_multiple_tables():
    tables = extract_tables("SELECT t1. FROM tabl1 t1, tabl2 t2")
    assert set(tables) == {(None, "tabl1", "t1", False), (None, "tabl2", "t2", False)}


def test_simple_insert_single_table():
    tables = extract_tables('insert into abc (id, name) values (1, "def")')

    # sqlparse mistakenly assigns an alias to the table
    # AND mistakenly identifies the field list as
    # assert tables == ((None, 'abc', 'abc', False),)

    assert tables == ((None, "abc", "abc", False),)


@pytest.mark.xfail
def test_simple_insert_single_table_schema_qualified():
    tables = extract_tables('insert into abc.def (id, name) values (1, "def")')
    assert tables == (("abc", "def", None, False),)


def test_simple_update_table_no_schema():
    tables = extract_tables("update abc set id = 1")
    assert tables == ((None, "abc", None, False),)


def test_simple_update_table_with_schema():
    tables = extract_tables("update abc.def set id = 1")
    assert tables == (("abc", "def", None, False),)


@pytest.mark.parametrize("join_type", ["", "INNER", "LEFT", "RIGHT OUTER"])
def test_join_table(join_type):
    sql = f"SELECT * FROM abc a {join_type} JOIN def d ON a.id = d.num"
    tables = extract_tables(sql)
    assert set(tables) == {(None, "abc", "a", False), (None, "def", "d", False)}


def test_join_table_schema_qualified():
    tables = extract_tables("SELECT * FROM abc.def x JOIN ghi.jkl y ON x.id = y.num")
    assert set(tables) == {("abc", "def", "x", False), ("ghi", "jkl", "y", False)}


def test_incomplete_join_clause():
    sql = """select a.x, b.y
             from abc a join bcd b
             on a.id = """
    tables = extract_tables(sql)
    assert tables == ((None, "abc", "a", False), (None, "bcd", "b", False))


def test_join_as_table():
    tables = extract_tables("SELECT * FROM my_table AS m WHERE m.a > 5")
    assert tables == ((None, "my_table", "m", False),)


def test_multiple_joins():
    sql = """select * from t1
            inner join t2 ON
              t1.id = t2.t1_id
            inner join t3 ON
              t2.id = t3."""
    tables = extract_tables(sql)
    assert tables == (
        (None, "t1", None, False),
        (None, "t2", None, False),
        (None, "t3", None, False),
    )


def test_subselect_tables():
    sql = "SELECT * FROM (SELECT  FROM abc"
    tables = extract_tables(sql)
    assert tables == ((None, "abc", None, False),)


@pytest.mark.parametrize("text", ["SELECT * FROM foo.", "SELECT 123 AS foo"])
def test_extract_no_tables(text):
    tables = extract_tables(text)
    assert tables == tuple()


@pytest.mark.parametrize("arg_list", ["", "arg1", "arg1, arg2, arg3"])
def test_simple_function_as_table(arg_list):
    tables = extract_tables(f"SELECT * FROM foo({arg_list})")
    assert tables == ((None, "foo", None, True),)


@pytest.mark.parametrize("arg_list", ["", "arg1", "arg1, arg2, arg3"])
def test_simple_schema_qualified_function_as_table(arg_list):
    tables = extract_tables(f"SELECT * FROM foo.bar({arg_list})")
    assert tables == (("foo", "bar", None, True),)


@pytest.mark.parametrize("arg_list", ["", "arg1", "arg1, arg2, arg3"])
def test_simple_aliased_function_as_table(arg_list):
    tables = extract_tables(f"SELECT * FROM foo({arg_list}) bar")
    assert tables == ((None, "foo", "bar", True),)


def test_simple_table_and_function():
    tables = extract_tables("SELECT * FROM foo JOIN bar()")
    assert set(tables) == {(None, "foo", None, False), (None, "bar", None, True)}


def test_complex_table_and_function():
    tables = extract_tables(
        """SELECT * FROM foo.bar baz
                               JOIN bar.qux(x, y, z) quux"""
    )
    assert set(tables) == {("foo", "bar", "baz", False), ("bar", "qux", "quux", True)}


def test_find_prev_keyword_using():
    q = "select * from tbl1 inner join tbl2 using (col1, "
    kw, q2 = find_prev_keyword(q)
    assert kw.value == "(" and q2 == "select * from tbl1 inner join tbl2 using ("


@pytest.mark.parametrize(
    "sql",
    [
        "select * from foo where bar",
        "select * from foo where bar = 1 and baz or ",
        "select * from foo where bar = 1 and baz between qux and ",
    ],
)
def test_find_prev_keyword_where(sql):
    kw, stripped = find_prev_keyword(sql)
    assert kw.value == "where" and stripped == "select * from foo where"


@pytest.mark.parametrize(
    "sql", ["create table foo (bar int, baz ", "select * from foo() as bar (baz "]
)
def test_find_prev_keyword_open_parens(sql):
    kw, _ = find_prev_keyword(sql)
    assert kw.value == "("


@pytest.mark.parametrize(
    "sql",
    [
        "",
        "$$ foo $$",
        "$$ 'foo' $$",
        '$$ "foo" $$',
        "$$ $a$ $$",
        "$a$ $$ $a$",
        "foo bar $$ baz $$",
    ],
)
def test_is_open_quote__closed(sql):
    assert not is_open_quote(sql)


@pytest.mark.parametrize(
    "sql",
    [
        "$$",
        ";;;$$",
        "foo $$ bar $$; foo $$",
        "$$ foo $a$",
        "foo 'bar baz",
        "$a$ foo ",
        '$$ "foo" ',
        "$$ $a$ ",
        "foo bar $$ baz",
    ],
)
def test_is_open_quote__open(sql):
    assert is_open_quote(sql)


@pytest.mark.parametrize(
    ("sql", "keywords", "expected"),
    [
        ("update abc set x = 1", ALL_KEYWORDS, True),
        ("update abc set x = 1 where y = 2", ALL_KEYWORDS, True),
        ("update abc set x = 1", BASE_KEYWORDS, True),
        ("update abc set x = 1 where y = 2", BASE_KEYWORDS, False),
        ("select x, y, z from abc", ALL_KEYWORDS, False),
        ("drop abc", ALL_KEYWORDS, True),
        ("alter abc", ALL_KEYWORDS, True),
        ("delete abc", ALL_KEYWORDS, True),
        ("truncate abc", ALL_KEYWORDS, True),
        ("insert into abc values (1, 2, 3)", ALL_KEYWORDS, False),
        ("insert into abc values (1, 2, 3)", BASE_KEYWORDS, False),
        ("insert into abc values (1, 2, 3)", ["insert"], True),
        ("insert into abc values (1, 2, 3)", ["insert"], True),
    ],
)
def test_is_destructive(sql, keywords, expected):
    assert is_destructive(sql, keywords) == expected


@pytest.mark.parametrize(
    ("warning_level", "expected"),
    [
        ("true", ALL_KEYWORDS),
        ("false", []),
        ("all", ALL_KEYWORDS),
        ("moderate", BASE_KEYWORDS),
        ("off", []),
        ("", []),
        (None, []),
        (ALL_KEYWORDS, ALL_KEYWORDS),
        (BASE_KEYWORDS, BASE_KEYWORDS),
        ("insert", ["insert"]),
        ("drop,alter,delete", ["drop", "alter", "delete"]),
        (["drop", "alter", "delete"], ["drop", "alter", "delete"]),
    ],
)
def test_parse_destructive_warning(warning_level, expected):
    assert parse_destructive_warning(warning_level) == expected



import copy
import os
import sys
import db_utils as dbutils
import fixture_utils as fixutils
import pexpect
import tempfile
import shutil
import signal


from steps import wrappers


def before_all(context):
    """Set env parameters."""
    env_old = copy.deepcopy(dict(os.environ))
    os.environ["LINES"] = "100"
    os.environ["COLUMNS"] = "100"
    os.environ["PAGER"] = "cat"
    os.environ["EDITOR"] = "ex"
    os.environ["VISUAL"] = "ex"
    os.environ["PROMPT_TOOLKIT_NO_CPR"] = "1"

    context.package_root = os.path.abspath(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    )
    fixture_dir = os.path.join(context.package_root, "tests/features/fixture_data")

    print("package root:", context.package_root)
    print("fixture dir:", fixture_dir)

    os.environ["COVERAGE_PROCESS_START"] = os.path.join(
        context.package_root, ".coveragerc"
    )

    context.exit_sent = False

    vi = "_".join([str(x) for x in sys.version_info[:3]])
    db_name = context.config.userdata.get("pg_test_db", "pgcli_behave_tests")
    db_name_full = f"{db_name}_{vi}"

    # Store get params from config.
    context.conf = {
        "host": context.config.userdata.get(
            "pg_test_host", os.getenv("PGHOST", "localhost")
        ),
        "user": context.config.userdata.get(
            "pg_test_user", os.getenv("PGUSER", "postgres")
        ),
        "pass": context.config.userdata.get(
            "pg_test_pass", os.getenv("PGPASSWORD", None)
        ),
        "port": context.config.userdata.get(
            "pg_test_port", os.getenv("PGPORT", "5432")
        ),
        "cli_command": (
            context.config.userdata.get("pg_cli_command", None)
            or '{python} -c "{startup}"'.format(
                python=sys.executable,
                startup="; ".join(
                    [
                        "import coverage",
                        "coverage.process_startup()",
                        "import pgcli.main",
                        "pgcli.main.cli(auto_envvar_prefix='BEHAVE')",
                    ]
                ),
            )
        ),
        "dbname": db_name_full,
        "dbname_tmp": db_name_full + "_tmp",
        "vi": vi,
        "pager_boundary": "---boundary---",
    }
    os.environ["PAGER"] = "{0} {1} {2}".format(
        sys.executable,
        os.path.join(context.package_root, "tests/features/wrappager.py"),
        context.conf["pager_boundary"],
    )

    # Store old env vars.
    context.pgenv = {
        "PGDATABASE": os.environ.get("PGDATABASE", None),
        "PGUSER": os.environ.get("PGUSER", None),
        "PGHOST": os.environ.get("PGHOST", None),
        "PGPASSWORD": os.environ.get("PGPASSWORD", None),
        "PGPORT": os.environ.get("PGPORT", None),
        "XDG_CONFIG_HOME": os.environ.get("XDG_CONFIG_HOME", None),
        "PGSERVICEFILE": os.environ.get("PGSERVICEFILE", None),
    }

    # Set new env vars.
    os.environ["PGDATABASE"] = context.conf["dbname"]
    os.environ["PGUSER"] = context.conf["user"]
    os.environ["PGHOST"] = context.conf["host"]
    os.environ["PGPORT"] = context.conf["port"]
    os.environ["PGSERVICEFILE"] = os.path.join(fixture_dir, "mock_pg_service.conf")

    if context.conf["pass"]:
        os.environ["PGPASSWORD"] = context.conf["pass"]
    else:
        if "PGPASSWORD" in os.environ:
            del os.environ["PGPASSWORD"]
    os.environ["BEHAVE_WARN"] = "moderate"

    context.cn = dbutils.create_db(
        context.conf["host"],
        context.conf["user"],
        context.conf["pass"],
        context.conf["dbname"],
        context.conf["port"],
    )
    context.pgbouncer_available = dbutils.pgbouncer_available(
        hostname=context.conf["host"],
        password=context.conf["pass"],
        username=context.conf["user"],
    )
    context.fixture_data = fixutils.read_fixture_files()

    # use temporary directory as config home
    context.env_config_home = tempfile.mkdtemp(prefix="pgcli_home_")
    os.environ["XDG_CONFIG_HOME"] = context.env_config_home
    show_env_changes(env_old, dict(os.environ))


def show_env_changes(env_old, env_new):
    """Print out all test-specific env values."""
    print("--- os.environ changed values: ---")
    all_keys = env_old.keys() | env_new.keys()
    for k in sorted(all_keys):
        old_value = env_old.get(k, "")
        new_value = env_new.get(k, "")
        if new_value and old_value != new_value:
            print(f'{k}="{new_value}"')
    print("-" * 20)


def after_all(context):
    """
    Unset env parameters.
    """
    dbutils.close_cn(context.cn)
    dbutils.drop_db(
        context.conf["host"],
        context.conf["user"],
        context.conf["pass"],
        context.conf["dbname"],
        context.conf["port"],
    )

    # Remove temp config directory
    shutil.rmtree(context.env_config_home)

    # Restore env vars.
    for k, v in context.pgenv.items():
        if k in os.environ and v is None:
            del os.environ[k]
        elif v:
            os.environ[k] = v


def before_step(context, _):
    context.atprompt = False


def is_known_problem(scenario):
    """TODO: why is this not working in 3.12?"""
    if sys.version_info >= (3, 12):
        return scenario.name in (
            'interrupt current query via "ctrl + c"',
            "run the cli with --username",
            "run the cli with --user",
            "run the cli with --port",
        )
    return False


def before_scenario(context, scenario):
    if scenario.name == "list databases":
        # not using the cli for that
        return
    if is_known_problem(scenario):
        scenario.skip()
    currentdb = None
    if "pgbouncer" in scenario.feature.tags:
        if context.pgbouncer_available:
            os.environ["PGDATABASE"] = "pgbouncer"
            os.environ["PGPORT"] = "6432"
            currentdb = "pgbouncer"
        else:
            scenario.skip()
    else:
        # set env vars back to normal test database
        os.environ["PGDATABASE"] = context.conf["dbname"]
        os.environ["PGPORT"] = context.conf["port"]
    wrappers.run_cli(context, currentdb=currentdb)
    wrappers.wait_prompt(context)


def after_scenario(context, scenario):
    """Cleans up after each scenario completes."""
    if hasattr(context, "cli") and context.cli and not context.exit_sent:
        # Quit nicely.
        if not getattr(context, "atprompt", False):
            dbname = context.currentdb
            context.cli.expect_exact(f"{dbname}>", timeout=5)
        try:
            context.cli.sendcontrol("c")
            context.cli.sendcontrol("d")
        except Exception as x:
            print("Failed cleanup after scenario:")
            print(x)
        try:
            context.cli.expect_exact(pexpect.EOF, timeout=5)
        except pexpect.TIMEOUT:
            print(f"--- after_scenario {scenario.name}: kill cli")
            context.cli.kill(signal.SIGKILL)
    if hasattr(context, "tmpfile_sql_help") and context.tmpfile_sql_help:
        context.tmpfile_sql_help.close()
        context.tmpfile_sql_help = None


# # TODO: uncomment to debug a failure
# def after_step(context, step):
#     if step.status == "failed":
#         import pdb; pdb.set_trace()



"""
Steps for behavioral style tests are defined in this module.
Each step is defined by the string decorating it.
This string is used to call the step in "*.feature" file.
"""

from behave import when, then
import wrappers


@when("we save a named query")
def step_save_named_query(context):
    """
    Send \ns command
    """
    context.cli.sendline("\\ns foo SELECT 12345")


@when("we use a named query")
def step_use_named_query(context):
    """
    Send \n command
    """
    context.cli.sendline("\\n foo")


@when("we delete a named query")
def step_delete_named_query(context):
    """
    Send \nd command
    """
    context.cli.sendline("\\nd foo")


@then("we see the named query saved")
def step_see_named_query_saved(context):
    """
    Wait to see query saved.
    """
    wrappers.expect_exact(context, "Saved.", timeout=2)


@then("we see the named query executed")
def step_see_named_query_executed(context):
    """
    Wait to see select output.
    """
    wrappers.expect_exact(context, "12345", timeout=1)
    wrappers.expect_exact(context, "SELECT 1", timeout=1)


@then("we see the named query deleted")
def step_see_named_query_deleted(context):
    """
    Wait to see query deleted.
    """
    wrappers.expect_pager(context, "foo: Deleted\r\n", timeout=1)



"""
Steps for behavioral style tests are defined in this module.
Each step is defined by the string decorating it.
This string is used to call the step in "*.feature" file.
"""

from behave import when, then
import wrappers


@when("we refresh completions")
def step_refresh_completions(context):
    """
    Send refresh command.
    """
    context.cli.sendline("\\refresh")


@then("we see completions refresh started")
def step_see_refresh_started(context):
    """
    Wait to see refresh output.
    """
    wrappers.expect_pager(
        context,
        [
            "Auto-completion refresh started in the background.\r\n",
            "Auto-completion refresh restarted.\r\n",
        ],
        timeout=2,
    )



import pytest


@pytest.fixture
def completer():
    import pgcli.pgcompleter as pgcompleter

    return pgcompleter.PGCompleter()


def test_ranking_ignores_identifier_quotes(completer):
    """When calculating result rank, identifier quotes should be ignored.

    The result ranking algorithm ignores identifier quotes. Without this
    correction, the match "user", which Postgres requires to be quoted
    since it is also a reserved word, would incorrectly fall below the
    match user_action because the literal quotation marks in "user"
    alter the position of the match.

    This test checks that the fuzzy ranking algorithm correctly ignores
    quotation marks when computing match ranks.

    """

    text = "user"
    collection = ["user_action", '"user"']
    matches = completer.find_matches(text, collection)
    assert len(matches) == 2


def test_ranking_based_on_shortest_match(completer):
    """Fuzzy result rank should be based on shortest match.

    Result ranking in fuzzy searching is partially based on the length
    of matches: shorter matches are considered more relevant than
    longer ones. When searching for the text 'user', the length
    component of the match 'user_group' could be either 4 ('user') or
    7 ('user_gr').

    This test checks that the fuzzy ranking algorithm uses the shorter
    match when calculating result rank.

    """

    text = "user"
    collection = ["api_user", "user_group"]
    matches = completer.find_matches(text, collection)

    assert matches[1].priority > matches[0].priority


@pytest.mark.parametrize(
    "collection",
    [["user_action", "user"], ["user_group", "user"], ["user_group", "user_action"]],
)
def test_should_break_ties_using_lexical_order(completer, collection):
    """Fuzzy result rank should use lexical order to break ties.

    When fuzzy matching, if multiple matches have the same match length and
    start position, present them in lexical (rather than arbitrary) order. For
    example, if we have tables 'user', 'user_action', and 'user_group', a
    search for the text 'user' should present these tables in this order.

    The input collections to this test are out of order; each run checks that
    the search text 'user' results in the input tables being reordered
    lexically.

    """

    text = "user"
    matches = completer.find_matches(text, collection)

    assert matches[1].priority > matches[0].priority


def test_matching_should_be_case_insensitive(completer):
    """Fuzzy matching should keep matches even if letter casing doesn't match.

    This test checks that variations of the text which have different casing
    are still matched.
    """

    text = "foo"
    collection = ["Foo", "FOO", "fOO"]
    matches = completer.find_matches(text, collection)

    assert len(matches) == 3



import errno
import shutil
import os
import platform
from os.path import expanduser, exists, dirname
import re
from typing import TextIO
from configobj import ConfigObj


def config_location():
    if "XDG_CONFIG_HOME" in os.environ:
        return "%s/pgcli/" % expanduser(os.environ["XDG_CONFIG_HOME"])
    elif platform.system() == "Windows":
        return os.getenv("USERPROFILE") + "\\AppData\\Local\\dbcli\\pgcli\\"
    else:
        return expanduser("~/.config/pgcli/")


def load_config(usr_cfg, def_cfg=None):
    # avoid config merges when possible. For writing, we need an umerged config instance.
    # see https://github.com/dbcli/pgcli/issues/1240 and https://github.com/DiffSK/configobj/issues/171
    if def_cfg:
        cfg = ConfigObj()
        cfg.merge(ConfigObj(def_cfg, interpolation=False))
        cfg.merge(ConfigObj(expanduser(usr_cfg), interpolation=False, encoding="utf-8"))
    else:
        cfg = ConfigObj(expanduser(usr_cfg), interpolation=False, encoding="utf-8")
    cfg.filename = expanduser(usr_cfg)
    return cfg


def ensure_dir_exists(path):
    parent_dir = expanduser(dirname(path))
    os.makedirs(parent_dir, exist_ok=True)


def write_default_config(source, destination, overwrite=False):
    destination = expanduser(destination)
    if not overwrite and exists(destination):
        return

    ensure_dir_exists(destination)

    shutil.copyfile(source, destination)


def upgrade_config(config, def_config):
    cfg = load_config(config, def_config)
    cfg.write()


def get_config_filename(pgclirc_file=None):
    return pgclirc_file or "%sconfig" % config_location()


def get_config(pgclirc_file=None):
    from pgcli import __file__ as package_root

    package_root = os.path.dirname(package_root)

    pgclirc_file = get_config_filename(pgclirc_file)

    default_config = os.path.join(package_root, "pgclirc")
    write_default_config(default_config, pgclirc_file)

    return load_config(pgclirc_file, default_config)


def get_casing_file(config):
    casing_file = config["main"]["casing_file"]
    if casing_file == "default":
        casing_file = config_location() + "casing"
    return casing_file


def skip_initial_comment(f_stream: TextIO) -> int:
    """
    Initial comment in ~/.pg_service.conf is not always marked with '#'
    which crashes the parser. This function takes a file object and
    "rewinds" it to the beginning of the first section,
    from where on it can be parsed safely

    :return: number of skipped lines
    """
    section_regex = r"\s*\["
    pos = f_stream.tell()
    lines_skipped = 0
    while True:
        line = f_stream.readline()
        if line == "":
            break
        if re.match(section_regex, line) is not None:
            f_stream.seek(pos)
            break
        else:
            pos += len(line)
            lines_skipped += 1
    return lines_skipped



# coding=utf-8

from pgcli.packages.parseutils.tables import extract_tables


supported_formats = (
    "sql-insert",
    "sql-update",
    "sql-update-1",
    "sql-update-2",
)

preprocessors = ()


def escape_for_sql_statement(value):
    if value is None:
        return "NULL"

    if isinstance(value, bytes):
        return f"X'{value.hex()}'"

    return "'{}'".format(value)


def adapter(data, headers, table_format=None, **kwargs):
    tables = extract_tables(formatter.query)
    if len(tables) > 0:
        table = tables[0]
        if table[0]:
            table_name = "{}.{}".format(*table[:2])
        else:
            table_name = table[1]
    else:
        table_name = "DUAL"
    if table_format == "sql-insert":
        h = '", "'.join(headers)
        yield 'INSERT INTO "{}" ("{}") VALUES'.format(table_name, h)
        prefix = "  "
        for d in data:
            values = ", ".join(escape_for_sql_statement(v) for i, v in enumerate(d))
            yield "{}({})".format(prefix, values)
            if prefix == "  ":
                prefix = ", "
        yield ";"
    if table_format.startswith("sql-update"):
        s = table_format.split("-")
        keys = 1
        if len(s) > 2:
            keys = int(s[-1])
        for d in data:
            yield 'UPDATE "{}" SET'.format(table_name)
            prefix = "  "
            for i, v in enumerate(d[keys:], keys):
                yield '{}"{}" = {}'.format(
                    prefix, headers[i], escape_for_sql_statement(v)
                )
                if prefix == "  ":
                    prefix = ", "
            f = '"{}" = {}'
            where = (
                f.format(headers[i], escape_for_sql_statement(d[i]))
                for i in range(keys)
            )
            yield "WHERE {};".format(" AND ".join(where))


def register_new_formatter(TabularOutputFormatter):
    global formatter
    formatter = TabularOutputFormatter
    for sql_format in supported_formats:
        TabularOutputFormatter.register_new_formatter(
            sql_format, adapter, preprocessors, {"table_format": sql_format}
        )



import re
import pexpect
from pgcli.main import COLOR_CODE_REGEX
import textwrap

from io import StringIO


def expect_exact(context, expected, timeout):
    timedout = False
    try:
        context.cli.expect_exact(expected, timeout=timeout)
    except pexpect.TIMEOUT:
        timedout = True
    if timedout:
        # Strip color codes out of the output.
        actual = re.sub(r"\x1b\[([0-9A-Za-z;?])+[m|K]?", "", context.cli.before)
        raise Exception(
            textwrap.dedent(
                """\
                Expected:
                ---
                {0!r}
                ---
                Actual:
                ---
                {1!r}
                ---
                Full log:
                ---
                {2!r}
                ---
            """
            ).format(expected, actual, context.logfile.getvalue())
        )


def expect_pager(context, expected, timeout):
    formatted = expected if isinstance(expected, list) else [expected]
    formatted = [
        f"{context.conf['pager_boundary']}\r\n{t}{context.conf['pager_boundary']}\r\n"
        for t in formatted
    ]

    expect_exact(
        context,
        formatted,
        timeout=timeout,
    )


def run_cli(context, run_args=None, prompt_check=True, currentdb=None):
    """Run the process using pexpect."""
    run_args = run_args or []
    cli_cmd = context.conf.get("cli_command")
    cmd_parts = [cli_cmd] + run_args
    cmd = " ".join(cmd_parts)
    context.cli = pexpect.spawnu(cmd, cwd=context.package_root)
    context.logfile = StringIO()
    context.cli.logfile = context.logfile
    context.exit_sent = False
    context.currentdb = currentdb or context.conf["dbname"]
    context.cli.sendline(r"\pset pager always")
    if prompt_check:
        wait_prompt(context)


def wait_prompt(context):
    """Make sure prompt is displayed."""
    prompt_str = "{0}>".format(context.currentdb)
    expect_exact(context, [prompt_str + " ", prompt_str, pexpect.EOF], timeout=3)



import pytest
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document
from utils import completions_to_set


@pytest.fixture
def completer():
    import pgcli.pgcompleter as pgcompleter

    return pgcompleter.PGCompleter(smart_completion=False)


@pytest.fixture
def complete_event():
    from unittest.mock import Mock

    return Mock()


def test_empty_string_completion(completer, complete_event):
    text = ""
    position = 0
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position), complete_event
        )
    )
    assert result == completions_to_set(map(Completion, completer.all_completions))


def test_select_keyword_completion(completer, complete_event):
    text = "SEL"
    position = len("SEL")
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position), complete_event
        )
    )
    assert result == completions_to_set([Completion(text="SELECT", start_position=-3)])


def test_function_name_completion(completer, complete_event):
    text = "SELECT MA"
    position = len("SELECT MA")
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position), complete_event
        )
    )
    assert result == completions_to_set(
        [
            Completion(text="MATERIALIZED VIEW", start_position=-2),
            Completion(text="MAX", start_position=-2),
            Completion(text="MAXEXTENTS", start_position=-2),
            Completion(text="MAKE_DATE", start_position=-2),
            Completion(text="MAKE_TIME", start_position=-2),
            Completion(text="MAKE_TIMESTAMPTZ", start_position=-2),
            Completion(text="MAKE_INTERVAL", start_position=-2),
            Completion(text="MASKLEN", start_position=-2),
            Completion(text="MAKE_TIMESTAMP", start_position=-2),
        ]
    )


def test_column_name_completion(completer, complete_event):
    text = "SELECT  FROM users"
    position = len("SELECT ")
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position), complete_event
        )
    )
    assert result == completions_to_set(map(Completion, completer.all_completions))


def test_alter_well_known_keywords_completion(completer, complete_event):
    text = "ALTER "
    position = len(text)
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position),
            complete_event,
            smart_completion=True,
        )
    )
    assert result > completions_to_set(
        [
            Completion(text="DATABASE", display_meta="keyword"),
            Completion(text="TABLE", display_meta="keyword"),
            Completion(text="SYSTEM", display_meta="keyword"),
        ]
    )
    assert (
        completions_to_set([Completion(text="CREATE", display_meta="keyword")])
        not in result
    )


def test_special_name_completion(completer, complete_event):
    text = "\\"
    position = len("\\")
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position), complete_event
        )
    )
    # Special commands will NOT be suggested during naive completion mode.
    assert result == completions_to_set([])


def test_datatype_name_completion(completer, complete_event):
    text = "SELECT price::IN"
    position = len("SELECT price::IN")
    result = completions_to_set(
        completer.get_completions(
            Document(text=text, cursor_position=position),
            complete_event,
            smart_completion=True,
        )
    )
    assert result == completions_to_set(
        [
            Completion(text="INET", display_meta="datatype"),
            Completion(text="INT", display_meta="datatype"),
            Completion(text="INT2", display_meta="datatype"),
            Completion(text="INT4", display_meta="datatype"),
            Completion(text="INT8", display_meta="datatype"),
            Completion(text="INTEGER", display_meta="datatype"),
            Completion(text="INTERNAL", display_meta="datatype"),
            Completion(text="INTERVAL", display_meta="datatype"),
        ]
    )



import re
import sqlparse
from sqlparse.tokens import Name
from collections import defaultdict
from .pgliterals.main import get_literals


white_space_regex = re.compile("\\s+", re.MULTILINE)


def _compile_regex(keyword):
    # Surround the keyword with word boundaries and replace interior whitespace
    # with whitespace wildcards
    pattern = "\\b" + white_space_regex.sub(r"\\s+", keyword) + "\\b"
    return re.compile(pattern, re.MULTILINE | re.IGNORECASE)


keywords = get_literals("keywords")
keyword_regexs = {kw: _compile_regex(kw) for kw in keywords}


class PrevalenceCounter:
    def __init__(self):
        self.keyword_counts = defaultdict(int)
        self.name_counts = defaultdict(int)

    def update(self, text):
        self.update_keywords(text)
        self.update_names(text)

    def update_names(self, text):
        for parsed in sqlparse.parse(text):
            for token in parsed.flatten():
                if token.ttype in Name:
                    self.name_counts[token.value] += 1

    def clear_names(self):
        self.name_counts = defaultdict(int)

    def update_keywords(self, text):
        # Count keywords. Can't rely for sqlparse for this, because it's
        # database agnostic
        for keyword, regex in keyword_regexs.items():
            for _ in regex.finditer(text):
                self.keyword_counts[keyword] += 1

    def keyword_count(self, keyword):
        return self.keyword_counts[keyword]

    def name_count(self, name):
        return self.name_counts[name]



"""
Steps for behavioral style tests are defined in this module.
Each step is defined by the string decorating it.
This string is used to call the step in "*.feature" file.
"""

import pexpect
import subprocess
import tempfile

from behave import when, then
from textwrap import dedent
import wrappers


@when("we list databases")
def step_list_databases(context):
    cmd = ["pgcli", "--list"]
    context.cmd_output = subprocess.check_output(cmd, cwd=context.package_root)


@then("we see list of databases")
def step_see_list_databases(context):
    assert b"List of databases" in context.cmd_output
    assert b"postgres" in context.cmd_output
    context.cmd_output = None


@when("we run dbcli")
def step_run_cli(context):
    wrappers.run_cli(context)


@when("we launch dbcli using {arg}")
def step_run_cli_using_arg(context, arg):
    prompt_check = False
    currentdb = None
    if arg == "--username":
        arg = "--username={}".format(context.conf["user"])
    if arg == "--user":
        arg = "--user={}".format(context.conf["user"])
    if arg == "--port":
        arg = "--port={}".format(context.conf["port"])
    if arg == "--password":
        arg = "--password"
        prompt_check = False
    # This uses the mock_pg_service.conf file in fixtures folder.
    if arg == "dsn_password":
        arg = "service=mock_postgres --password"
        prompt_check = False
        currentdb = "postgres"
    wrappers.run_cli(
        context, run_args=[arg], prompt_check=prompt_check, currentdb=currentdb
    )


@when("we wait for prompt")
def step_wait_prompt(context):
    wrappers.wait_prompt(context)


@when('we send "ctrl + d"')
def step_ctrl_d(context):
    """
    Send Ctrl + D to hopefully exit.
    """
    step_try_to_ctrl_d(context)
    context.cli.expect(pexpect.EOF, timeout=5)
    context.exit_sent = True


@when('we try to send "ctrl + d"')
def step_try_to_ctrl_d(context):
    """
    Send Ctrl + D, perhaps exiting, perhaps not (if a transaction is
    ongoing).
    """
    # turn off pager before exiting
    context.cli.sendcontrol("c")
    context.cli.sendline(r"\pset pager off")
    wrappers.wait_prompt(context)
    context.cli.sendcontrol("d")


@when('we send "ctrl + c"')
def step_ctrl_c(context):
    """Send Ctrl + c to hopefully interrupt."""
    context.cli.sendcontrol("c")


@then("we see cancelled query warning")
def step_see_cancelled_query_warning(context):
    """
    Make sure we receive the warning that the current query was cancelled.
    """
    wrappers.expect_exact(context, "cancelled query", timeout=2)


@then("we see ongoing transaction message")
def step_see_ongoing_transaction_error(context):
    """
    Make sure we receive the warning that a transaction is ongoing.
    """
    context.cli.expect("A transaction is ongoing.", timeout=2)


@when("we send sleep query")
def step_send_sleep_15_seconds(context):
    """
    Send query to sleep for 15 seconds.
    """
    context.cli.sendline("select pg_sleep(15)")


@when("we check for any non-idle sleep queries")
def step_check_for_active_sleep_queries(context):
    """
    Send query to check for any non-idle pg_sleep queries.
    """
    context.cli.sendline(
        "select state from pg_stat_activity where query not like '%pg_stat_activity%' and query like '%pg_sleep%' and state != 'idle';"
    )


@then("we don't see any non-idle sleep queries")
def step_no_active_sleep_queries(context):
    """Confirm that any pg_sleep queries are either idle or not active."""
    wrappers.expect_exact(
        context,
        context.conf["pager_boundary"]
        + "\r"
        + dedent(
            """
            +-------+\r
            | state |\r
            |-------|\r
            +-------+\r
            SELECT 0\r
        """
        )
        + context.conf["pager_boundary"],
        timeout=5,
    )


@when(r'we send "\?" command')
def step_send_help(context):
    r"""
    Send \? to see help.
    """
    context.cli.sendline(r"\?")


@when("we send partial select command")
def step_send_partial_select_command(context):
    """
    Send `SELECT a` to see completion.
    """
    context.cli.sendline("SELECT a")


@then("we see error message")
def step_see_error_message(context):
    wrappers.expect_exact(context, 'column "a" does not exist', timeout=2)


@when("we send source command")
def step_send_source_command(context):
    context.tmpfile_sql_help = tempfile.NamedTemporaryFile(prefix="pgcli_")
    context.tmpfile_sql_help.write(rb"\?")
    context.tmpfile_sql_help.flush()
    context.cli.sendline(rf"\i {context.tmpfile_sql_help.name}")
    wrappers.expect_exact(context, context.conf["pager_boundary"] + "\r\n", timeout=5)


@when("we run query to check application_name")
def step_check_application_name(context):
    context.cli.sendline(
        "SELECT 'found' FROM pg_stat_activity WHERE application_name = 'pgcli' HAVING COUNT(*) > 0;"
    )


@then("we see found")
def step_see_found(context):
    wrappers.expect_exact(
        context,
        context.conf["pager_boundary"]
        + "\r"
        + dedent(
            """
            +----------+\r
            | ?column? |\r
            |----------|\r
            | found    |\r
            +----------+\r
            SELECT 1\r
        """
        )
        + context.conf["pager_boundary"],
        timeout=5,
    )


@then("we respond to the destructive warning: {response}")
def step_resppond_to_destructive_command(context, response):
    """Respond to destructive command."""
    wrappers.expect_exact(
        context,
        "You're about to run a destructive command.\r\nDo you want to proceed? [y/N]:",
        timeout=2,
    )
    context.cli.sendline(response.strip())


@then("we send password")
def step_send_password(context):
    wrappers.expect_exact(context, "Password for", timeout=5)
    context.cli.sendline(context.conf["pass"] or "DOES NOT MATTER")


@when('we send "{text}"')
def step_send_text(context, text):
    context.cli.sendline(text)
    # Try to detect whether we are exiting. If so, set `exit_sent`
    # so that `after_scenario` correctly cleans up.
    try:
        context.cli.expect(pexpect.EOF, timeout=0.2)
    except pexpect.TIMEOUT:
        pass
    else:
        context.exit_sent = True



import pytest
from sqlparse import parse
from pgcli.packages.parseutils.ctes import (
    token_start_pos,
    extract_ctes,
    extract_column_names as _extract_column_names,
)


def extract_column_names(sql):
    p = parse(sql)[0]
    return _extract_column_names(p)


def test_token_str_pos():
    sql = "SELECT * FROM xxx"
    p = parse(sql)[0]
    idx = p.token_index(p.tokens[-1])
    assert token_start_pos(p.tokens, idx) == len("SELECT * FROM ")

    sql = "SELECT * FROM \nxxx"
    p = parse(sql)[0]
    idx = p.token_index(p.tokens[-1])
    assert token_start_pos(p.tokens, idx) == len("SELECT * FROM \n")


def test_single_column_name_extraction():
    sql = "SELECT abc FROM xxx"
    assert extract_column_names(sql) == ("abc",)


def test_aliased_single_column_name_extraction():
    sql = "SELECT abc def FROM xxx"
    assert extract_column_names(sql) == ("def",)


def test_aliased_expression_name_extraction():
    sql = "SELECT 99 abc FROM xxx"
    assert extract_column_names(sql) == ("abc",)


def test_multiple_column_name_extraction():
    sql = "SELECT abc, def FROM xxx"
    assert extract_column_names(sql) == ("abc", "def")


def test_missing_column_name_handled_gracefully():
    sql = "SELECT abc, 99 FROM xxx"
    assert extract_column_names(sql) == ("abc",)

    sql = "SELECT abc, 99, def FROM xxx"
    assert extract_column_names(sql) == ("abc", "def")


def test_aliased_multiple_column_name_extraction():
    sql = "SELECT abc def, ghi jkl FROM xxx"
    assert extract_column_names(sql) == ("def", "jkl")


def test_table_qualified_column_name_extraction():
    sql = "SELECT abc.def, ghi.jkl FROM xxx"
    assert extract_column_names(sql) == ("def", "jkl")


@pytest.mark.parametrize(
    "sql",
    [
        "INSERT INTO foo (x, y, z) VALUES (5, 6, 7) RETURNING x, y",
        "DELETE FROM foo WHERE x > y RETURNING x, y",
        "UPDATE foo SET x = 9 RETURNING x, y",
    ],
)
def test_extract_column_names_from_returning_clause(sql):
    assert extract_column_names(sql) == ("x", "y")


def test_simple_cte_extraction():
    sql = "WITH a AS (SELECT abc FROM xxx) SELECT * FROM a"
    start_pos = len("WITH a AS ")
    stop_pos = len("WITH a AS (SELECT abc FROM xxx)")
    ctes, remainder = extract_ctes(sql)

    assert tuple(ctes) == (("a", ("abc",), start_pos, stop_pos),)
    assert remainder.strip() == "SELECT * FROM a"


def test_cte_extraction_around_comments():
    sql = """--blah blah blah
            WITH a AS (SELECT abc def FROM x)
            SELECT * FROM a"""
    start_pos = len(
        """--blah blah blah
            WITH a AS """
    )
    stop_pos = len(
        """--blah blah blah
            WITH a AS (SELECT abc def FROM x)"""
    )

    ctes, remainder = extract_ctes(sql)
    assert tuple(ctes) == (("a", ("def",), start_pos, stop_pos),)
    assert remainder.strip() == "SELECT * FROM a"


def test_multiple_cte_extraction():
    sql = """WITH
            x AS (SELECT abc, def FROM x),
            y AS (SELECT ghi, jkl FROM y)
            SELECT * FROM a, b"""

    start1 = len(
        """WITH
            x AS """
    )

    stop1 = len(
        """WITH
            x AS (SELECT abc, def FROM x)"""
    )

    start2 = len(
        """WITH
            x AS (SELECT abc, def FROM x),
            y AS """
    )

    stop2 = len(
        """WITH
            x AS (SELECT abc, def FROM x),
            y AS (SELECT ghi, jkl FROM y)"""
    )

    ctes, remainder = extract_ctes(sql)
    assert tuple(ctes) == (
        ("x", ("abc", "def"), start1, stop1),
        ("y", ("ghi", "jkl"), start2, stop2),
    )



import platform
from setuptools import setup, find_packages

from pgcli import __version__

description = "CLI for Postgres Database. With auto-completion and syntax highlighting."

install_requirements = [
    "pgspecial>=2.0.0",
    "click >= 4.1",
    "Pygments>=2.0",  # Pygments has to be Capitalcased. WTF?
    # We still need to use pt-2 unless pt-3 released on Fedora32
    # see: https://github.com/dbcli/pgcli/pull/1197
    "prompt_toolkit>=2.0.6,<4.0.0",
    "psycopg >= 3.0.14",
    "sqlparse >=0.3.0,<0.5",
    "configobj >= 5.0.6",
    "pendulum>=2.1.0",
    "cli_helpers[styles] >= 2.2.1",
]


# setproctitle is used to mask the password when running `ps` in command line.
# But this is not necessary in Windows since the password is never shown in the
# task manager. Also setproctitle is a hard dependency to install in Windows,
# so we'll only install it if we're not in Windows.
if platform.system() != "Windows" and not platform.system().startswith("CYGWIN"):
    install_requirements.append("setproctitle >= 1.1.9")

# Windows will require the binary psycopg to run pgcli
if platform.system() == "Windows":
    install_requirements.append("psycopg-binary >= 3.0.14")


setup(
    name="pgcli",
    author="Pgcli Core Team",
    author_email="pgcli-dev@googlegroups.com",
    version=__version__,
    license="BSD",
    url="http://pgcli.com",
    packages=find_packages(),
    package_data={"pgcli": ["pgclirc", "packages/pgliterals/pgliterals.json"]},
    description=description,
    long_description=open("README.rst").read(),
    install_requires=install_requirements,
    dependency_links=[
        "http://github.com/psycopg/repo/tarball/master#egg=psycopg-3.0.10"
    ],
    extras_require={
        "keyring": ["keyring >= 12.2.0"],
        "sshtunnel": ["sshtunnel >= 0.4.0"],
    },
    python_requires=">=3.8",
    entry_points="""
        [console_scripts]
        pgcli=pgcli.main:cli
    """,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: SQL",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import os
import os.path

from behave import when, then
import wrappers


@when("we start external editor providing a file name")
def step_edit_file(context):
    """Edit file with external editor."""
    context.editor_file_name = os.path.join(
        context.package_root, "test_file_{0}.sql".format(context.conf["vi"])
    )
    if os.path.exists(context.editor_file_name):
        os.remove(context.editor_file_name)
    context.cli.sendline(r"\e {}".format(os.path.basename(context.editor_file_name)))
    wrappers.expect_exact(
        context, 'Entering Ex mode.  Type "visual" to go to Normal mode.', timeout=2
    )
    wrappers.expect_exact(context, ":", timeout=2)


@when("we type sql in the editor")
def step_edit_type_sql(context):
    context.cli.sendline("i")
    context.cli.sendline("select * from abc")
    context.cli.sendline(".")
    wrappers.expect_exact(context, ":", timeout=2)


@when("we exit the editor")
def step_edit_quit(context):
    context.cli.sendline("x")
    wrappers.expect_exact(context, "written", timeout=2)


@then("we see the sql in prompt")
def step_edit_done_sql(context):
    for match in "select * from abc".split(" "):
        wrappers.expect_exact(context, match, timeout=1)
    # Cleanup the command line.
    context.cli.sendcontrol("c")
    # Cleanup the edited file.
    if context.editor_file_name and os.path.exists(context.editor_file_name):
        os.remove(context.editor_file_name)
    context.atprompt = True


@when("we tee output")
def step_tee_ouptut(context):
    context.tee_file_name = os.path.join(
        context.package_root, "tee_file_{0}.sql".format(context.conf["vi"])
    )
    if os.path.exists(context.tee_file_name):
        os.remove(context.tee_file_name)
    context.cli.sendline(r"\o {}".format(os.path.basename(context.tee_file_name)))
    wrappers.expect_exact(context, context.conf["pager_boundary"] + "\r\n", timeout=5)
    wrappers.expect_exact(context, "Writing to file", timeout=5)
    wrappers.expect_exact(context, context.conf["pager_boundary"] + "\r\n", timeout=5)
    wrappers.expect_exact(context, "Time", timeout=5)


@when('we query "select 123456"')
def step_query_select_123456(context):
    context.cli.sendline("select 123456")


@when("we stop teeing output")
def step_notee_output(context):
    context.cli.sendline(r"\o")
    wrappers.expect_exact(context, "Time", timeout=5)


@then("we see 123456 in tee output")
def step_see_123456_in_ouput(context):
    with open(context.tee_file_name) as f:
        assert "123456" in f.read()
    if os.path.exists(context.tee_file_name):
        os.remove(context.tee_file_name)
    context.atprompt = True



#!/usr/bin/env python
"""A script to publish a release of pgcli to PyPI."""

import io
from optparse import OptionParser
import re
import subprocess
import sys

import click

DEBUG = False
CONFIRM_STEPS = False
DRY_RUN = False


def skip_step():
    """
    Asks for user's response whether to run a step. Default is yes.
    :return: boolean
    """
    global CONFIRM_STEPS

    if CONFIRM_STEPS:
        return not click.confirm("--- Run this step?", default=True)
    return False


def run_step(*args):
    """
    Prints out the command and asks if it should be run.
    If yes (default), runs it.
    :param args: list of strings (command and args)
    """
    global DRY_RUN

    cmd = args
    print(" ".join(cmd))
    if skip_step():
        print("--- Skipping...")
    elif DRY_RUN:
        print("--- Pretending to run...")
    else:
        subprocess.check_output(cmd)


def version(version_file):
    _version_re = re.compile(
        r'__version__\s+=\s+(?P<quote>[\'"])(?P<version>.*)(?P=quote)'
    )

    with io.open(version_file, encoding="utf-8") as f:
        ver = _version_re.search(f.read()).group("version")

    return ver


def commit_for_release(version_file, ver):
    run_step("git", "reset")
    run_step("git", "add", "-u")
    run_step("git", "commit", "--message", "Releasing version {}".format(ver))


def create_git_tag(tag_name):
    run_step("git", "tag", tag_name)


def create_distribution_files():
    run_step("python", "setup.py", "clean", "--all", "sdist", "bdist_wheel")


def upload_distribution_files():
    run_step("twine", "upload", "dist/*")


def push_to_github():
    run_step("git", "push", "origin", "main")


def push_tags_to_github():
    run_step("git", "push", "--tags", "origin")


def checklist(questions):
    for question in questions:
        if not click.confirm("--- {}".format(question), default=False):
            sys.exit(1)


if __name__ == "__main__":
    if DEBUG:
        subprocess.check_output = lambda x: x

    checks = [
        "Have you updated the AUTHORS file?",
        "Have you updated the `Usage` section of the README?",
    ]
    checklist(checks)

    ver = version("pgcli/__init__.py")
    print("Releasing Version:", ver)

    parser = OptionParser()
    parser.add_option(
        "-c",
        "--confirm-steps",
        action="store_true",
        dest="confirm_steps",
        default=False,
        help=(
            "Confirm every step. If the step is not " "confirmed, it will be skipped."
        ),
    )
    parser.add_option(
        "-d",
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Print out, but not actually run any steps.",
    )

    popts, pargs = parser.parse_args()
    CONFIRM_STEPS = popts.confirm_steps
    DRY_RUN = popts.dry_run

    if not click.confirm("Are you sure?", default=False):
        sys.exit(1)

    commit_for_release("pgcli/__init__.py", ver)
    create_git_tag("v{}".format(ver))
    create_distribution_files()
    push_to_github()
    push_tags_to_github()
    upload_distribution_files()



import os
from unittest.mock import patch, MagicMock, ANY

import pytest
from configobj import ConfigObj
from click.testing import CliRunner
from sshtunnel import SSHTunnelForwarder

from pgcli.main import cli, PGCli
from pgcli.pgexecute import PGExecute


@pytest.fixture
def mock_ssh_tunnel_forwarder() -> MagicMock:
    mock_ssh_tunnel_forwarder = MagicMock(
        SSHTunnelForwarder, local_bind_ports=[1111], autospec=True
    )
    with patch(
        "pgcli.main.sshtunnel.SSHTunnelForwarder",
        return_value=mock_ssh_tunnel_forwarder,
    ) as mock:
        yield mock


@pytest.fixture
def mock_pgexecute() -> MagicMock:
    with patch.object(PGExecute, "__init__", return_value=None) as mock_pgexecute:
        yield mock_pgexecute


def test_ssh_tunnel(
    mock_ssh_tunnel_forwarder: MagicMock, mock_pgexecute: MagicMock
) -> None:
    # Test with just a host
    tunnel_url = "some.host"
    db_params = {
        "database": "dbname",
        "host": "db.host",
        "user": "db_user",
        "passwd": "db_passwd",
    }
    expected_tunnel_params = {
        "local_bind_address": ("127.0.0.1",),
        "remote_bind_address": (db_params["host"], 5432),
        "ssh_address_or_host": (tunnel_url, 22),
        "logger": ANY,
    }

    pgcli = PGCli(ssh_tunnel_url=tunnel_url)
    pgcli.connect(**db_params)

    mock_ssh_tunnel_forwarder.assert_called_once_with(**expected_tunnel_params)
    mock_ssh_tunnel_forwarder.return_value.start.assert_called_once()
    mock_pgexecute.assert_called_once()

    call_args, call_kwargs = mock_pgexecute.call_args
    assert call_args == (
        db_params["database"],
        db_params["user"],
        db_params["passwd"],
        "127.0.0.1",
        pgcli.ssh_tunnel.local_bind_ports[0],
        "",
    )
    mock_ssh_tunnel_forwarder.reset_mock()
    mock_pgexecute.reset_mock()

    # Test with a full url and with a specific db port
    tunnel_user = "tunnel_user"
    tunnel_passwd = "tunnel_pass"
    tunnel_host = "some.other.host"
    tunnel_port = 1022
    tunnel_url = f"ssh://{tunnel_user}:{tunnel_passwd}@{tunnel_host}:{tunnel_port}"
    db_params["port"] = 1234

    expected_tunnel_params["remote_bind_address"] = (
        db_params["host"],
        db_params["port"],
    )
    expected_tunnel_params["ssh_address_or_host"] = (tunnel_host, tunnel_port)
    expected_tunnel_params["ssh_username"] = tunnel_user
    expected_tunnel_params["ssh_password"] = tunnel_passwd

    pgcli = PGCli(ssh_tunnel_url=tunnel_url)
    pgcli.connect(**db_params)

    mock_ssh_tunnel_forwarder.assert_called_once_with(**expected_tunnel_params)
    mock_ssh_tunnel_forwarder.return_value.start.assert_called_once()
    mock_pgexecute.assert_called_once()

    call_args, call_kwargs = mock_pgexecute.call_args
    assert call_args == (
        db_params["database"],
        db_params["user"],
        db_params["passwd"],
        "127.0.0.1",
        pgcli.ssh_tunnel.local_bind_ports[0],
        "",
    )
    mock_ssh_tunnel_forwarder.reset_mock()
    mock_pgexecute.reset_mock()

    # Test with DSN
    dsn = (
        f"user={db_params['user']} password={db_params['passwd']} "
        f"host={db_params['host']} port={db_params['port']}"
    )

    pgcli = PGCli(ssh_tunnel_url=tunnel_url)
    pgcli.connect(dsn=dsn)

    expected_dsn = (
        f"user={db_params['user']} password={db_params['passwd']} "
        f"host=127.0.0.1 port={pgcli.ssh_tunnel.local_bind_ports[0]}"
    )

    mock_ssh_tunnel_forwarder.assert_called_once_with(**expected_tunnel_params)
    mock_pgexecute.assert_called_once()

    call_args, call_kwargs = mock_pgexecute.call_args
    assert expected_dsn in call_args


def test_cli_with_tunnel() -> None:
    runner = CliRunner()
    tunnel_url = "mytunnel"
    with patch.object(
        PGCli, "__init__", autospec=True, return_value=None
    ) as mock_pgcli:
        runner.invoke(cli, ["--ssh-tunnel", tunnel_url])
        mock_pgcli.assert_called_once()
        call_args, call_kwargs = mock_pgcli.call_args
        assert call_kwargs["ssh_tunnel_url"] == tunnel_url


def test_config(
    tmpdir: os.PathLike, mock_ssh_tunnel_forwarder: MagicMock, mock_pgexecute: MagicMock
) -> None:
    pgclirc = str(tmpdir.join("rcfile"))

    tunnel_user = "tunnel_user"
    tunnel_passwd = "tunnel_pass"
    tunnel_host = "tunnel.host"
    tunnel_port = 1022
    tunnel_url = f"{tunnel_user}:{tunnel_passwd}@{tunnel_host}:{tunnel_port}"

    tunnel2_url = "tunnel2.host"

    config = ConfigObj()
    config.filename = pgclirc
    config["ssh tunnels"] = {}
    config["ssh tunnels"][r"\.com$"] = tunnel_url
    config["ssh tunnels"][r"^hello-"] = tunnel2_url
    config.write()

    # Unmatched host
    pgcli = PGCli(pgclirc_file=pgclirc)
    pgcli.connect(host="unmatched.host")
    mock_ssh_tunnel_forwarder.assert_not_called()

    # Host matching first tunnel
    pgcli = PGCli(pgclirc_file=pgclirc)
    pgcli.connect(host="matched.host.com")
    mock_ssh_tunnel_forwarder.assert_called_once()
    call_args, call_kwargs = mock_ssh_tunnel_forwarder.call_args
    assert call_kwargs["ssh_address_or_host"] == (tunnel_host, tunnel_port)
    assert call_kwargs["ssh_username"] == tunnel_user
    assert call_kwargs["ssh_password"] == tunnel_passwd
    mock_ssh_tunnel_forwarder.reset_mock()

    # Host matching second tunnel
    pgcli = PGCli(pgclirc_file=pgclirc)
    pgcli.connect(host="hello-i-am-matched")
    mock_ssh_tunnel_forwarder.assert_called_once()

    call_args, call_kwargs = mock_ssh_tunnel_forwarder.call_args
    assert call_kwargs["ssh_address_or_host"] == (tunnel2_url, 22)
    mock_ssh_tunnel_forwarder.reset_mock()

    # Host matching both tunnels (will use the first one matched)
    pgcli = PGCli(pgclirc_file=pgclirc)
    pgcli.connect(host="hello-i-am-matched.com")
    mock_ssh_tunnel_forwarder.assert_called_once()

    call_args, call_kwargs = mock_ssh_tunnel_forwarder.call_args
    assert call_kwargs["ssh_address_or_host"] == (tunnel_host, tunnel_port)
    assert call_kwargs["ssh_username"] == tunnel_user
    assert call_kwargs["ssh_password"] == tunnel_passwd



import logging
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.filters import (
    completion_is_selected,
    is_searching,
    has_completions,
    has_selection,
    vi_mode,
)

from .pgbuffer import buffer_should_be_handled, safe_multi_line_mode

_logger = logging.getLogger(__name__)


def pgcli_bindings(pgcli):
    """Custom key bindings for pgcli."""
    kb = KeyBindings()

    tab_insert_text = " " * 4

    @kb.add("f2")
    def _(event):
        """Enable/Disable SmartCompletion Mode."""
        _logger.debug("Detected F2 key.")
        pgcli.completer.smart_completion = not pgcli.completer.smart_completion

    @kb.add("f3")
    def _(event):
        """Enable/Disable Multiline Mode."""
        _logger.debug("Detected F3 key.")
        pgcli.multi_line = not pgcli.multi_line

    @kb.add("f4")
    def _(event):
        """Toggle between Vi and Emacs mode."""
        _logger.debug("Detected F4 key.")
        pgcli.vi_mode = not pgcli.vi_mode
        event.app.editing_mode = EditingMode.VI if pgcli.vi_mode else EditingMode.EMACS

    @kb.add("f5")
    def _(event):
        """Toggle between Vi and Emacs mode."""
        _logger.debug("Detected F5 key.")
        pgcli.explain_mode = not pgcli.explain_mode

    @kb.add("tab")
    def _(event):
        """Force autocompletion at cursor on non-empty lines."""

        _logger.debug("Detected <Tab> key.")

        buff = event.app.current_buffer
        doc = buff.document

        if doc.on_first_line or doc.current_line.strip():
            if buff.complete_state:
                buff.complete_next()
            else:
                buff.start_completion(select_first=True)
        else:
            buff.insert_text(tab_insert_text, fire_event=False)

    @kb.add("escape", filter=has_completions)
    def _(event):
        """Force closing of autocompletion."""
        _logger.debug("Detected <Esc> key.")

        event.current_buffer.complete_state = None
        event.app.current_buffer.complete_state = None

    @kb.add("c-space")
    def _(event):
        """
        Initialize autocompletion at cursor.

        If the autocompletion menu is not showing, display it with the
        appropriate completions for the context.

        If the menu is showing, select the next completion.
        """
        _logger.debug("Detected <C-Space> key.")

        b = event.app.current_buffer
        if b.complete_state:
            b.complete_next()
        else:
            b.start_completion(select_first=False)

    @kb.add("enter", filter=completion_is_selected)
    def _(event):
        """Makes the enter key work as the tab key only when showing the menu.

        In other words, don't execute query when enter is pressed in
        the completion dropdown menu, instead close the dropdown menu
        (accept current selection).

        """
        _logger.debug("Detected enter key during completion selection.")

        event.current_buffer.complete_state = None
        event.app.current_buffer.complete_state = None

    # When using multi_line input mode the buffer is not handled on Enter (a new line is
    # inserted instead), so we force the handling if we're not in a completion or
    # history search, and one of several conditions are True
    @kb.add(
        "enter",
        filter=~(completion_is_selected | is_searching)
        & buffer_should_be_handled(pgcli),
    )
    def _(event):
        _logger.debug("Detected enter key.")
        event.current_buffer.validate_and_handle()

    @kb.add("escape", "enter", filter=~vi_mode & ~safe_multi_line_mode(pgcli))
    def _(event):
        """Introduces a line break regardless of multi-line mode or not."""
        _logger.debug("Detected alt-enter key.")
        event.app.current_buffer.insert_text("\n")

    @kb.add("c-p", filter=~has_selection)
    def _(event):
        """Move up in history."""
        event.current_buffer.history_backward(count=event.arg)

    @kb.add("c-n", filter=~has_selection)
    def _(event):
        """Move down in history."""
        event.current_buffer.history_forward(count=event.arg)

    return kb



from textwrap import dedent

import psycopg
import pytest
from unittest.mock import patch, MagicMock
from pgspecial.main import PGSpecial, NO_QUERY
from utils import run, dbtest, requires_json, requires_jsonb

from pgcli.main import PGCli
from pgcli.packages.parseutils.meta import FunctionMetadata


def function_meta_data(
    func_name,
    schema_name="public",
    arg_names=None,
    arg_types=None,
    arg_modes=None,
    return_type=None,
    is_aggregate=False,
    is_window=False,
    is_set_returning=False,
    is_extension=False,
    arg_defaults=None,
):
    return FunctionMetadata(
        schema_name,
        func_name,
        arg_names,
        arg_types,
        arg_modes,
        return_type,
        is_aggregate,
        is_window,
        is_set_returning,
        is_extension,
        arg_defaults,
    )


@dbtest
def test_conn(executor):
    run(executor, """create table test(a text)""")
    run(executor, """insert into test values('abc')""")
    assert run(executor, """select * from test""", join=True) == dedent(
        """\
        +-----+
        | a   |
        |-----|
        | abc |
        +-----+
        SELECT 1"""
    )


@dbtest
def test_copy(executor):
    executor_copy = executor.copy()
    run(executor_copy, """create table test(a text)""")
    run(executor_copy, """insert into test values('abc')""")
    assert run(executor_copy, """select * from test""", join=True) == dedent(
        """\
        +-----+
        | a   |
        |-----|
        | abc |
        +-----+
        SELECT 1"""
    )


@dbtest
def test_bools_are_treated_as_strings(executor):
    run(executor, """create table test(a boolean)""")
    run(executor, """insert into test values(True)""")
    assert run(executor, """select * from test""", join=True) == dedent(
        """\
        +------+
        | a    |
        |------|
        | True |
        +------+
        SELECT 1"""
    )


@dbtest
def test_expanded_slash_G(executor, pgspecial):
    # Tests whether we reset the expanded output after a \G.
    run(executor, """create table test(a boolean)""")
    run(executor, """insert into test values(True)""")
    results = run(executor, r"""select * from test \G""", pgspecial=pgspecial)
    assert pgspecial.expanded_output == False


@dbtest
def test_schemata_table_views_and_columns_query(executor):
    run(executor, "create table a(x text, y text)")
    run(executor, "create table b(z text)")
    run(executor, "create view d as select 1 as e")
    run(executor, "create schema schema1")
    run(executor, "create table schema1.c (w text DEFAULT 'meow')")
    run(executor, "create schema schema2")

    # schemata
    # don't enforce all members of the schemas since they may include postgres
    # temporary schemas
    assert set(executor.schemata()) >= {
        "public",
        "pg_catalog",
        "information_schema",
        "schema1",
        "schema2",
    }
    assert executor.search_path() == ["pg_catalog", "public"]

    # tables
    assert set(executor.tables()) >= {
        ("public", "a"),
        ("public", "b"),
        ("schema1", "c"),
    }

    assert set(executor.table_columns()) >= {
        ("public", "a", "x", "text", False, None),
        ("public", "a", "y", "text", False, None),
        ("public", "b", "z", "text", False, None),
        ("schema1", "c", "w", "text", True, "'meow'::text"),
    }

    # views
    assert set(executor.views()) >= {("public", "d")}

    assert set(executor.view_columns()) >= {
        ("public", "d", "e", "integer", False, None)
    }


@dbtest
def test_foreign_key_query(executor):
    run(executor, "create schema schema1")
    run(executor, "create schema schema2")
    run(executor, "create table schema1.parent(parentid int PRIMARY KEY)")
    run(
        executor,
        "create table schema2.child(childid int PRIMARY KEY, motherid int REFERENCES schema1.parent)",
    )

    assert set(executor.foreignkeys()) >= {
        ("schema1", "parent", "parentid", "schema2", "child", "motherid")
    }


@dbtest
def test_functions_query(executor):
    run(
        executor,
        """create function func1() returns int
                     language sql as $$select 1$$""",
    )
    run(executor, "create schema schema1")
    run(
        executor,
        """create function schema1.func2() returns int
                     language sql as $$select 2$$""",
    )

    run(
        executor,
        """create function func3()
                     returns table(x int, y int) language sql
                     as $$select 1, 2 from generate_series(1,5)$$;""",
    )

    run(
        executor,
        """create function func4(x int) returns setof int language sql
                     as $$select generate_series(1,5)$$;""",
    )

    funcs = set(executor.functions())
    assert funcs >= {
        function_meta_data(func_name="func1", return_type="integer"),
        function_meta_data(
            func_name="func3",
            arg_names=["x", "y"],
            arg_types=["integer", "integer"],
            arg_modes=["t", "t"],
            return_type="record",
            is_set_returning=True,
        ),
        function_meta_data(
            schema_name="public",
            func_name="func4",
            arg_names=("x",),
            arg_types=("integer",),
            return_type="integer",
            is_set_returning=True,
        ),
        function_meta_data(
            schema_name="schema1", func_name="func2", return_type="integer"
        ),
    }


@dbtest
def test_datatypes_query(executor):
    run(executor, "create type foo AS (a int, b text)")

    types = list(executor.datatypes())
    assert types == [("public", "foo")]


@dbtest
def test_database_list(executor):
    databases = executor.databases()
    assert "_test_db" in databases


@dbtest
def test_invalid_syntax(executor, exception_formatter):
    result = run(executor, "invalid syntax!", exception_formatter=exception_formatter)
    assert 'syntax error at or near "invalid"' in result[0]


@dbtest
def test_invalid_column_name(executor, exception_formatter):
    result = run(
        executor, "select invalid command", exception_formatter=exception_formatter
    )
    assert 'column "invalid" does not exist' in result[0]


@pytest.fixture(params=[True, False])
def expanded(request):
    return request.param


@dbtest
def test_unicode_support_in_output(executor, expanded):
    run(executor, "create table unicodechars(t text)")
    run(executor, "insert into unicodechars (t) values ('é')")

    # See issue #24, this raises an exception without proper handling
    assert "é" in run(
        executor, "select * from unicodechars", join=True, expanded=expanded
    )


@dbtest
def test_not_is_special(executor, pgspecial):
    """is_special is set to false for database queries."""
    query = "select 1"
    result = list(executor.run(query, pgspecial=pgspecial))
    success, is_special = result[0][5:]
    assert success == True
    assert is_special == False


@dbtest
def test_execute_from_file_no_arg(executor, pgspecial):
    r"""\i without a filename returns an error."""
    result = list(executor.run(r"\i", pgspecial=pgspecial))
    status, sql, success, is_special = result[0][3:]
    assert "missing required argument" in status
    assert success == False
    assert is_special == True


@dbtest
@patch("pgcli.main.os")
def test_execute_from_file_io_error(os, executor, pgspecial):
    r"""\i with an os_error returns an error."""
    # Inject an OSError.
    os.path.expanduser.side_effect = OSError("test")

    # Check the result.
    result = list(executor.run(r"\i test", pgspecial=pgspecial))
    status, sql, success, is_special = result[0][3:]
    assert status == "test"
    assert success == False
    assert is_special == True


@dbtest
def test_execute_from_commented_file_that_executes_another_file(
    executor, pgspecial, tmpdir
):
    # https://github.com/dbcli/pgcli/issues/1336
    sqlfile1 = tmpdir.join("test01.sql")
    sqlfile1.write("-- asdf \n\\h")
    sqlfile2 = tmpdir.join("test00.sql")
    sqlfile2.write("--An useless comment;\nselect now();\n-- another useless comment")

    rcfile = str(tmpdir.join("rcfile"))
    print(rcfile)
    cli = PGCli(pgexecute=executor, pgclirc_file=rcfile)
    assert cli != None
    statement = "--comment\n\\h"
    result = run(executor, statement, pgspecial=cli.pgspecial)
    assert result != None
    assert result[0].find("ALTER TABLE")


@dbtest
def test_execute_commented_first_line_and_special(executor, pgspecial, tmpdir):
    # just some base cases that should work also
    statement = "--comment\nselect now();"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("now") >= 0

    statement = "/*comment*/\nselect now();"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("now") >= 0

    # https://github.com/dbcli/pgcli/issues/1362
    statement = "--comment\n\\h"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = "--comment1\n--comment2\n\\h"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = "/*comment*/\n\h;"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = """/*comment1
    comment2*/
    \h"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = """/*comment1
    comment2*/
    /*comment 3
    comment4*/
    \\h"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = "    /*comment*/\n\h;"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = "/*comment\ncomment line2*/\n\h;"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = "          /*comment\ncomment line2*/\n\h;"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("ALTER") >= 0
    assert result[1].find("ABORT") >= 0

    statement = """\\h /*comment4 */"""
    result = run(executor, statement, pgspecial=pgspecial)
    print(result)
    assert result != None
    assert result[0].find("No help") >= 0

    # TODO: we probably don't want to do this but sqlparse is not parsing things well
    # we relly want it to find help but right now, sqlparse isn't dropping the /*comment*/
    # style comments after command

    statement = """/*comment1*/
    \h
    /*comment4 */"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[0].find("No help") >= 0

    # TODO: same for this one
    statement = """/*comment1
    comment3
    comment2*/
    \\h
    /*comment4
    comment5
    comment6*/"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[0].find("No help") >= 0


@dbtest
def test_execute_commented_first_line_and_normal(executor, pgspecial, tmpdir):
    # https://github.com/dbcli/pgcli/issues/1403

    # just some base cases that should work also
    statement = "--comment\nselect now();"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("now") >= 0

    statement = "/*comment*/\nselect now();"
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[1].find("now") >= 0

    # this simulates the original error (1403) without having to add/drop tables
    # since it was just an error on reading input files and not the actual
    # command itself

    # test that the statement works
    statement = """VALUES (1, 'one'), (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # test the statement with a \n in the middle
    statement = """VALUES (1, 'one'),\n (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # test the statement with a newline in the middle
    statement = """VALUES (1, 'one'),
     (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # now add a single comment line
    statement = """--comment\nVALUES (1, 'one'), (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # doing without special char \n
    statement = """--comment
    VALUES (1,'one'),
    (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # two comment lines
    statement = """--comment\n--comment2\nVALUES (1,'one'), (2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # doing without special char \n
    statement = """--comment
    --comment2
    VALUES (1,'one'), (2, 'two'), (3, 'three');
    """
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # multiline comment + newline in middle of the statement
    statement = """/*comment
comment2
comment3*/
VALUES (1,'one'),
(2, 'two'), (3, 'three');"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0

    # multiline comment + newline in middle of the statement
    # + comments after the statement
    statement = """/*comment
comment2
comment3*/
VALUES (1,'one'),
(2, 'two'), (3, 'three');
--comment4
--comment5"""
    result = run(executor, statement, pgspecial=pgspecial)
    assert result != None
    assert result[5].find("three") >= 0


@dbtest
def test_multiple_queries_same_line(executor):
    result = run(executor, "select 'foo'; select 'bar'")
    assert len(result) == 12  # 2 * (output+status) * 3 lines
    assert "foo" in result[3]
    assert "bar" in result[9]


@dbtest
def test_multiple_queries_with_special_command_same_line(executor, pgspecial):
    result = run(executor, r"select 'foo'; \d", pgspecial=pgspecial)
    assert len(result) == 11  # 2 * (output+status) * 3 lines
    assert "foo" in result[3]
    # This is a lame check. :(
    assert "Schema" in result[7]


@dbtest
def test_multiple_queries_same_line_syntaxerror(executor, exception_formatter):
    result = run(
        executor,
        "select 'fooé'; invalid syntax é",
        exception_formatter=exception_formatter,
    )
    assert "fooé" in result[3]
    assert 'syntax error at or near "invalid"' in result[-1]


@pytest.fixture
def pgspecial():
    return PGCli().pgspecial


@dbtest
def test_special_command_help(executor, pgspecial):
    result = run(executor, "\\?", pgspecial=pgspecial)[1].split("|")
    assert "Command" in result[1]
    assert "Description" in result[2]


@dbtest
def test_bytea_field_support_in_output(executor):
    run(executor, "create table binarydata(c bytea)")
    run(executor, "insert into binarydata (c) values (decode('DEADBEEF', 'hex'))")

    assert "\\xdeadbeef" in run(executor, "select * from binarydata", join=True)


@dbtest
def test_unicode_support_in_unknown_type(executor):
    assert "日本語" in run(executor, "SELECT '日本語' AS japanese;", join=True)


@dbtest
def test_unicode_support_in_enum_type(executor):
    run(executor, "CREATE TYPE mood AS ENUM ('sad', 'ok', 'happy', '日本語')")
    run(executor, "CREATE TABLE person (name TEXT, current_mood mood)")
    run(executor, "INSERT INTO person VALUES ('Moe', '日本語')")
    assert "日本語" in run(executor, "SELECT * FROM person", join=True)


@requires_json
def test_json_renders_without_u_prefix(executor, expanded):
    run(executor, "create table jsontest(d json)")
    run(executor, """insert into jsontest (d) values ('{"name": "Éowyn"}')""")
    result = run(
        executor, "SELECT d FROM jsontest LIMIT 1", join=True, expanded=expanded
    )

    assert '{"name": "Éowyn"}' in result


@requires_jsonb
def test_jsonb_renders_without_u_prefix(executor, expanded):
    run(executor, "create table jsonbtest(d jsonb)")
    run(executor, """insert into jsonbtest (d) values ('{"name": "Éowyn"}')""")
    result = run(
        executor, "SELECT d FROM jsonbtest LIMIT 1", join=True, expanded=expanded
    )

    assert '{"name": "Éowyn"}' in result


@dbtest
def test_date_time_types(executor):
    run(executor, "SET TIME ZONE UTC")
    assert (
        run(executor, "SELECT (CAST('00:00:00' AS time))", join=True).split("\n")[3]
        == "| 00:00:00 |"
    )
    assert (
        run(executor, "SELECT (CAST('00:00:00+14:59' AS timetz))", join=True).split(
            "\n"
        )[3]
        == "| 00:00:00+14:59 |"
    )
    assert (
        run(executor, "SELECT (CAST('4713-01-01 BC' AS date))", join=True).split("\n")[
            3
        ]
        == "| 4713-01-01 BC |"
    )
    assert (
        run(
            executor, "SELECT (CAST('4713-01-01 00:00:00 BC' AS timestamp))", join=True
        ).split("\n")[3]
        == "| 4713-01-01 00:00:00 BC |"
    )
    assert (
        run(
            executor,
            "SELECT (CAST('4713-01-01 00:00:00+00 BC' AS timestamptz))",
            join=True,
        ).split("\n")[3]
        == "| 4713-01-01 00:00:00+00 BC |"
    )
    assert (
        run(
            executor, "SELECT (CAST('-123456789 days 12:23:56' AS interval))", join=True
        ).split("\n")[3]
        == "| -123456789 days, 12:23:56 |"
    )


@dbtest
@pytest.mark.parametrize("value", ["10000000", "10000000.0", "10000000000000"])
def test_large_numbers_render_directly(executor, value):
    run(executor, "create table numbertest(a numeric)")
    run(executor, f"insert into numbertest (a) values ({value})")
    assert value in run(executor, "select * from numbertest", join=True)


@dbtest
@pytest.mark.parametrize("command", ["di", "dv", "ds", "df", "dT"])
@pytest.mark.parametrize("verbose", ["", "+"])
@pytest.mark.parametrize("pattern", ["", "x", "*.*", "x.y", "x.*", "*.y"])
def test_describe_special(executor, command, verbose, pattern, pgspecial):
    # We don't have any tests for the output of any of the special commands,
    # but we can at least make sure they run without error
    sql = r"\{command}{verbose} {pattern}".format(**locals())
    list(executor.run(sql, pgspecial=pgspecial))


@dbtest
@pytest.mark.parametrize("sql", ["invalid sql", "SELECT 1; select error;"])
def test_raises_with_no_formatter(executor, sql):
    with pytest.raises(psycopg.ProgrammingError):
        list(executor.run(sql))


@dbtest
def test_on_error_resume(executor, exception_formatter):
    sql = "select 1; error; select 1;"
    result = list(
        executor.run(sql, on_error_resume=True, exception_formatter=exception_formatter)
    )
    assert len(result) == 3


@dbtest
def test_on_error_stop(executor, exception_formatter):
    sql = "select 1; error; select 1;"
    result = list(
        executor.run(
            sql, on_error_resume=False, exception_formatter=exception_formatter
        )
    )
    assert len(result) == 2


# @dbtest
# def test_unicode_notices(executor):
#     sql = "DO language plpgsql $$ BEGIN RAISE NOTICE '有人更改'; END $$;"
#     result = list(executor.run(sql))
#     assert result[0][0] == u'NOTICE:  有人更改\n'


@dbtest
def test_nonexistent_function_definition(executor):
    with pytest.raises(RuntimeError):
        result = executor.view_definition("there_is_no_such_function")


@dbtest
def test_function_definition(executor):
    run(
        executor,
        """
            CREATE OR REPLACE FUNCTION public.the_number_three()
            RETURNS int
            LANGUAGE sql
            AS $function$
              select 3;
            $function$
    """,
    )
    result = executor.function_definition("the_number_three")


@dbtest
def test_view_definition(executor):
    run(executor, "create table tbl1 (a text, b numeric)")
    run(executor, "create view vw1 AS SELECT * FROM tbl1")
    run(executor, "create materialized view mvw1 AS SELECT * FROM tbl1")
    result = executor.view_definition("vw1")
    assert 'VIEW "public"."vw1" AS' in result
    assert "FROM tbl1" in result
    # import pytest; pytest.set_trace()
    result = executor.view_definition("mvw1")
    assert "MATERIALIZED VIEW" in result


@dbtest
def test_nonexistent_view_definition(executor):
    with pytest.raises(RuntimeError):
        result = executor.view_definition("there_is_no_such_view")
    with pytest.raises(RuntimeError):
        result = executor.view_definition("mvw1")


@dbtest
def test_short_host(executor):
    with patch.object(executor, "host", "localhost"):
        assert executor.short_host == "localhost"
    with patch.object(executor, "host", "localhost.example.org"):
        assert executor.short_host == "localhost"
    with patch.object(
        executor, "host", "localhost1.example.org,localhost2.example.org"
    ):
        assert executor.short_host == "localhost1"


class VirtualCursor:
    """Mock a cursor to virtual database like pgbouncer."""

    def __init__(self):
        self.protocol_error = False
        self.protocol_message = ""
        self.description = None
        self.status = None
        self.statusmessage = "Error"

    def execute(self, *args, **kwargs):
        self.protocol_error = True
        self.protocol_message = "Command not supported"


@dbtest
def test_exit_without_active_connection(executor):
    quit_handler = MagicMock()
    pgspecial = PGSpecial()
    pgspecial.register(
        quit_handler,
        "\\q",
        "\\q",
        "Quit pgcli.",
        arg_type=NO_QUERY,
        case_sensitive=True,
        aliases=(":q",),
    )

    with patch.object(
        executor.conn, "cursor", side_effect=psycopg.InterfaceError("I'm broken!")
    ):
        # we should be able to quit the app, even without active connection
        run(executor, "\\q", pgspecial=pgspecial)
        quit_handler.assert_called_once()

        # an exception should be raised when running a query without active connection
        with pytest.raises(psycopg.InterfaceError):
            run(executor, "select 1", pgspecial=pgspecial)


@dbtest
def test_virtual_database(executor):
    virtual_connection = MagicMock()
    virtual_connection.cursor.return_value = VirtualCursor()
    with patch.object(executor, "conn", virtual_connection):
        result = run(executor, "select 1")
        assert "Command not supported" in result



from metadata import (
    MetaData,
    alias,
    name_join,
    fk_join,
    join,
    keyword,
    schema,
    table,
    view,
    function,
    column,
    wildcard_expansion,
    get_result,
    result_set,
    qual,
    no_qual,
    parametrize,
)
from prompt_toolkit.completion import Completion
from utils import completions_to_set


metadata = {
    "tables": {
        "users": ["id", "parentid", "email", "first_name", "last_name"],
        "Users": ["userid", "username"],
        "orders": ["id", "ordered_date", "status", "email"],
        "select": ["id", "insert", "ABC"],
    },
    "views": {"user_emails": ["id", "email"], "functions": ["function"]},
    "functions": [
        ["custom_fun", [], [], [], "", False, False, False, False],
        ["_custom_fun", [], [], [], "", False, False, False, False],
        ["custom_func1", [], [], [], "", False, False, False, False],
        ["custom_func2", [], [], [], "", False, False, False, False],
        [
            "set_returning_func",
            ["x", "y"],
            ["integer", "integer"],
            ["b", "b"],
            "",
            False,
            False,
            True,
            False,
        ],
    ],
    "datatypes": ["custom_type1", "custom_type2"],
    "foreignkeys": [
        ("public", "users", "id", "public", "users", "parentid"),
        ("public", "users", "id", "public", "Users", "userid"),
    ],
}

metadata = {k: {"public": v} for k, v in metadata.items()}

testdata = MetaData(metadata)

cased_users_col_names = ["ID", "PARENTID", "Email", "First_Name", "last_name"]
cased_users2_col_names = ["UserID", "UserName"]
cased_func_names = [
    "Custom_Fun",
    "_custom_fun",
    "Custom_Func1",
    "custom_func2",
    "set_returning_func",
]
cased_tbls = ["Users", "Orders"]
cased_views = ["User_Emails", "Functions"]
casing = (
    ["SELECT", "PUBLIC"]
    + cased_func_names
    + cased_tbls
    + cased_views
    + cased_users_col_names
    + cased_users2_col_names
)
# Lists for use in assertions
cased_funcs = [
    function(f)
    for f in ("Custom_Fun()", "_custom_fun()", "Custom_Func1()", "custom_func2()")
] + [function("set_returning_func(x := , y := )", display="set_returning_func(x, y)")]
cased_tbls = [table(t) for t in (cased_tbls + ['"Users"', '"select"'])]
cased_rels = [view(t) for t in cased_views] + cased_funcs + cased_tbls
cased_users_cols = [column(c) for c in cased_users_col_names]
aliased_rels = (
    [table(t) for t in ("users u", '"Users" U', "orders o", '"select" s')]
    + [view("user_emails ue"), view("functions f")]
    + [
        function(f)
        for f in (
            "_custom_fun() cf",
            "custom_fun() cf",
            "custom_func1() cf",
            "custom_func2() cf",
        )
    ]
    + [
        function(
            "set_returning_func(x := , y := ) srf",
            display="set_returning_func(x, y) srf",
        )
    ]
)
cased_aliased_rels = (
    [table(t) for t in ("Users U", '"Users" U', "Orders O", '"select" s')]
    + [view("User_Emails UE"), view("Functions F")]
    + [
        function(f)
        for f in (
            "_custom_fun() cf",
            "Custom_Fun() CF",
            "Custom_Func1() CF",
            "custom_func2() cf",
        )
    ]
    + [
        function(
            "set_returning_func(x := , y := ) srf",
            display="set_returning_func(x, y) srf",
        )
    ]
)
completers = testdata.get_completers(casing)


# Just to make sure that this doesn't crash
@parametrize("completer", completers())
def test_function_column_name(completer):
    for l in range(
        len("SELECT * FROM Functions WHERE function:"),
        len("SELECT * FROM Functions WHERE function:text") + 1,
    ):
        assert [] == get_result(
            completer, "SELECT * FROM Functions WHERE function:text"[:l]
        )


@parametrize("action", ["ALTER", "DROP", "CREATE", "CREATE OR REPLACE"])
@parametrize("completer", completers())
def test_drop_alter_function(completer, action):
    assert get_result(completer, action + " FUNCTION set_ret") == [
        function("set_returning_func(x integer, y integer)", -len("set_ret"))
    ]


@parametrize("completer", completers())
def test_empty_string_completion(completer):
    result = get_result(completer, "")
    assert completions_to_set(
        testdata.keywords() + testdata.specials()
    ) == completions_to_set(result)


@parametrize("completer", completers())
def test_select_keyword_completion(completer):
    result = get_result(completer, "SEL")
    assert completions_to_set(result) == completions_to_set([keyword("SELECT", -3)])


@parametrize("completer", completers())
def test_builtin_function_name_completion(completer):
    result = get_result(completer, "SELECT MA")
    assert completions_to_set(result) == completions_to_set(
        [
            function("MAKE_DATE", -2),
            function("MAKE_INTERVAL", -2),
            function("MAKE_TIME", -2),
            function("MAKE_TIMESTAMP", -2),
            function("MAKE_TIMESTAMPTZ", -2),
            function("MASKLEN", -2),
            function("MAX", -2),
            keyword("MAXEXTENTS", -2),
            keyword("MATERIALIZED VIEW", -2),
        ]
    )


@parametrize("completer", completers())
def test_builtin_function_matches_only_at_start(completer):
    text = "SELECT IN"

    result = [c.text for c in get_result(completer, text)]

    assert "MIN" not in result


@parametrize("completer", completers(casing=False, aliasing=False))
def test_user_function_name_completion(completer):
    result = get_result(completer, "SELECT cu")
    assert completions_to_set(result) == completions_to_set(
        [
            function("custom_fun()", -2),
            function("_custom_fun()", -2),
            function("custom_func1()", -2),
            function("custom_func2()", -2),
            function("CURRENT_DATE", -2),
            function("CURRENT_TIMESTAMP", -2),
            function("CUME_DIST", -2),
            function("CURRENT_TIME", -2),
            keyword("CURRENT", -2),
        ]
    )


@parametrize("completer", completers(casing=False, aliasing=False))
def test_user_function_name_completion_matches_anywhere(completer):
    result = get_result(completer, "SELECT om")
    assert completions_to_set(result) == completions_to_set(
        [
            function("custom_fun()", -2),
            function("_custom_fun()", -2),
            function("custom_func1()", -2),
            function("custom_func2()", -2),
        ]
    )


@parametrize("completer", completers(casing=True))
def test_list_functions_for_special(completer):
    result = get_result(completer, r"\df ")
    assert completions_to_set(result) == completions_to_set(
        [schema("PUBLIC")] + [function(f) for f in cased_func_names]
    )


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_suggested_column_names_from_visible_table(completer):
    result = get_result(completer, "SELECT  from users", len("SELECT "))
    assert completions_to_set(result) == completions_to_set(
        testdata.columns_functions_and_keywords("users")
    )


@parametrize("completer", completers(casing=True, qualify=no_qual))
def test_suggested_cased_column_names(completer):
    result = get_result(completer, "SELECT  from users", len("SELECT "))
    assert completions_to_set(result) == completions_to_set(
        cased_funcs
        + cased_users_cols
        + testdata.builtin_functions()
        + testdata.keywords()
    )


@parametrize("completer", completers(casing=False, qualify=no_qual))
@parametrize("text", ["SELECT  from users", "INSERT INTO Orders SELECT  from users"])
def test_suggested_auto_qualified_column_names(text, completer):
    position = text.index("  ") + 1
    cols = [column(c.lower()) for c in cased_users_col_names]
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        cols + testdata.functions_and_keywords()
    )


@parametrize("completer", completers(casing=False, qualify=qual))
@parametrize(
    "text",
    [
        'SELECT  from users U NATURAL JOIN "Users"',
        'INSERT INTO Orders SELECT  from users U NATURAL JOIN "Users"',
    ],
)
def test_suggested_auto_qualified_column_names_two_tables(text, completer):
    position = text.index("  ") + 1
    cols = [column("U." + c.lower()) for c in cased_users_col_names]
    cols += [column('"Users".' + c.lower()) for c in cased_users2_col_names]
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        cols + testdata.functions_and_keywords()
    )


@parametrize("completer", completers(casing=True, qualify=["always"]))
@parametrize("text", ["UPDATE users SET ", "INSERT INTO users("])
def test_no_column_qualification(text, completer):
    cols = [column(c) for c in cased_users_col_names]
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(cols)


@parametrize("completer", completers(casing=True, qualify=["always"]))
def test_suggested_cased_always_qualified_column_names(completer):
    text = "SELECT  from users"
    position = len("SELECT ")
    cols = [column("users." + c) for c in cased_users_col_names]
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        cased_funcs + cols + testdata.builtin_functions() + testdata.keywords()
    )


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_suggested_column_names_in_function(completer):
    result = get_result(completer, "SELECT MAX( from users", len("SELECT MAX("))
    assert completions_to_set(result) == completions_to_set(
        testdata.columns_functions_and_keywords("users")
    )


@parametrize("completer", completers(casing=False))
def test_suggested_column_names_with_table_dot(completer):
    result = get_result(completer, "SELECT users. from users", len("SELECT users."))
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=False))
def test_suggested_column_names_with_alias(completer):
    result = get_result(completer, "SELECT u. from users u", len("SELECT u."))
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_suggested_multiple_column_names(completer):
    result = get_result(completer, "SELECT id,  from users u", len("SELECT id, "))
    assert completions_to_set(result) == completions_to_set(
        testdata.columns_functions_and_keywords("users")
    )


@parametrize("completer", completers(casing=False))
def test_suggested_multiple_column_names_with_alias(completer):
    result = get_result(
        completer, "SELECT u.id, u. from users u", len("SELECT u.id, u.")
    )
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=True))
def test_suggested_cased_column_names_with_alias(completer):
    result = get_result(
        completer, "SELECT u.id, u. from users u", len("SELECT u.id, u.")
    )
    assert completions_to_set(result) == completions_to_set(cased_users_cols)


@parametrize("completer", completers(casing=False))
def test_suggested_multiple_column_names_with_dot(completer):
    result = get_result(
        completer,
        "SELECT users.id, users. from users u",
        len("SELECT users.id, users."),
    )
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=False))
def test_suggest_columns_after_three_way_join(completer):
    text = """SELECT * FROM users u1
              INNER JOIN users u2 ON u1.id = u2.id
              INNER JOIN users u3 ON u2.id = u3."""
    result = get_result(completer, text)
    assert column("id") in result


join_condition_texts = [
    'INSERT INTO orders SELECT * FROM users U JOIN "Users" U2 ON ',
    """INSERT INTO public.orders(orderid)
    SELECT * FROM users U JOIN "Users" U2 ON """,
    'SELECT * FROM users U JOIN "Users" U2 ON ',
    'SELECT * FROM users U INNER join "Users" U2 ON ',
    'SELECT * FROM USERS U right JOIN "Users" U2 ON ',
    'SELECT * FROM users U LEFT JOIN "Users" U2 ON ',
    'SELECT * FROM Users U FULL JOIN "Users" U2 ON ',
    'SELECT * FROM users U right outer join "Users" U2 ON ',
    'SELECT * FROM Users U LEFT OUTER JOIN "Users" U2 ON ',
    'SELECT * FROM users U FULL OUTER JOIN "Users" U2 ON ',
    """SELECT *
    FROM users U
    FULL OUTER JOIN "Users" U2 ON
    """,
]


@parametrize("completer", completers(casing=False))
@parametrize("text", join_condition_texts)
def test_suggested_join_conditions(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [alias("U"), alias("U2"), fk_join("U2.userid = U.id")]
    )


@parametrize("completer", completers(casing=True))
@parametrize("text", join_condition_texts)
def test_cased_join_conditions(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [alias("U"), alias("U2"), fk_join("U2.UserID = U.ID")]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        """SELECT *
    FROM users
    CROSS JOIN "Users"
    NATURAL JOIN users u
    JOIN "Users" u2 ON
    """
    ],
)
def test_suggested_join_conditions_with_same_table_twice(completer, text):
    result = get_result(completer, text)
    assert result == [
        fk_join("u2.userid = u.id"),
        fk_join("u2.userid = users.id"),
        name_join('u2.userid = "Users".userid'),
        name_join('u2.username = "Users".username'),
        alias("u"),
        alias("u2"),
        alias("users"),
        alias('"Users"'),
    ]


@parametrize("completer", completers())
@parametrize("text", ["SELECT * FROM users JOIN users u2 on foo."])
def test_suggested_join_conditions_with_invalid_qualifier(completer, text):
    result = get_result(completer, text)
    assert result == []


@parametrize("completer", completers(casing=False))
@parametrize(
    ("text", "ref"),
    [
        ("SELECT * FROM users JOIN NonTable on ", "NonTable"),
        ("SELECT * FROM users JOIN nontable nt on ", "nt"),
    ],
)
def test_suggested_join_conditions_with_invalid_table(completer, text, ref):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [alias("users"), alias(ref)]
    )


@parametrize("completer", completers(casing=False, aliasing=False))
@parametrize(
    "text",
    [
        'SELECT * FROM "Users" u JOIN u',
        'SELECT * FROM "Users" u JOIN uid',
        'SELECT * FROM "Users" u JOIN userid',
        'SELECT * FROM "Users" u JOIN id',
    ],
)
def test_suggested_joins_fuzzy(completer, text):
    result = get_result(completer, text)
    last_word = text.split()[-1]
    expected = join("users ON users.id = u.userid", -len(last_word))
    assert expected in result


join_texts = [
    "SELECT * FROM Users JOIN ",
    """INSERT INTO "Users"
    SELECT *
    FROM Users
    INNER JOIN """,
    """INSERT INTO public."Users"(username)
    SELECT *
    FROM Users
    INNER JOIN """,
    """SELECT *
    FROM Users
    INNER JOIN """,
]


@parametrize("completer", completers(casing=False, aliasing=False))
@parametrize("text", join_texts)
def test_suggested_joins(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas_and_from_clause_items()
        + [
            join('"Users" ON "Users".userid = Users.id'),
            join("users users2 ON users2.id = Users.parentid"),
            join("users users2 ON users2.parentid = Users.id"),
        ]
    )


@parametrize("completer", completers(casing=True, aliasing=False))
@parametrize("text", join_texts)
def test_cased_joins(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [schema("PUBLIC")]
        + cased_rels
        + [
            join('"Users" ON "Users".UserID = Users.ID'),
            join("Users Users2 ON Users2.ID = Users.PARENTID"),
            join("Users Users2 ON Users2.PARENTID = Users.ID"),
        ]
    )


@parametrize("completer", completers(casing=False, aliasing=True))
@parametrize("text", join_texts)
def test_aliased_joins(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas()
        + aliased_rels
        + [
            join('"Users" U ON U.userid = Users.id'),
            join("users u ON u.id = Users.parentid"),
            join("users u ON u.parentid = Users.id"),
        ]
    )


@parametrize("completer", completers(casing=False, aliasing=False))
@parametrize(
    "text",
    [
        'SELECT * FROM public."Users" JOIN ',
        'SELECT * FROM public."Users" RIGHT OUTER JOIN ',
        """SELECT *
    FROM public."Users"
    LEFT JOIN """,
    ],
)
def test_suggested_joins_quoted_schema_qualified_table(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas_and_from_clause_items()
        + [join('public.users ON users.id = "Users".userid')]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT u.name, o.id FROM users u JOIN orders o ON ",
        "SELECT u.name, o.id FROM users u JOIN orders o ON JOIN orders o2 ON",
    ],
)
def test_suggested_aliases_after_on(completer, text):
    position = len("SELECT u.name, o.id FROM users u JOIN orders o ON ")
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        [
            alias("u"),
            name_join("o.id = u.id"),
            name_join("o.email = u.email"),
            alias("o"),
        ]
    )


@parametrize("completer", completers())
@parametrize(
    "text",
    [
        "SELECT u.name, o.id FROM users u JOIN orders o ON o.user_id = ",
        "SELECT u.name, o.id FROM users u JOIN orders o ON o.user_id =  JOIN orders o2 ON",
    ],
)
def test_suggested_aliases_after_on_right_side(completer, text):
    position = len("SELECT u.name, o.id FROM users u JOIN orders o ON o.user_id = ")
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set([alias("u"), alias("o")])


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT users.name, orders.id FROM users JOIN orders ON ",
        "SELECT users.name, orders.id FROM users JOIN orders ON JOIN orders orders2 ON",
    ],
)
def test_suggested_tables_after_on(completer, text):
    position = len("SELECT users.name, orders.id FROM users JOIN orders ON ")
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        [
            name_join("orders.id = users.id"),
            name_join("orders.email = users.email"),
            alias("users"),
            alias("orders"),
        ]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT users.name, orders.id FROM users JOIN orders ON orders.user_id = JOIN orders orders2 ON",
        "SELECT users.name, orders.id FROM users JOIN orders ON orders.user_id = ",
    ],
)
def test_suggested_tables_after_on_right_side(completer, text):
    position = len(
        "SELECT users.name, orders.id FROM users JOIN orders ON orders.user_id = "
    )
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        [alias("users"), alias("orders")]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT * FROM users INNER JOIN orders USING (",
        "SELECT * FROM users INNER JOIN orders USING(",
    ],
)
def test_join_using_suggests_common_columns(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [column("id"), column("email")]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT * FROM users u1 JOIN users u2 USING (email) JOIN user_emails ue USING()",
        "SELECT * FROM users u1 JOIN users u2 USING(email) JOIN user_emails ue USING ()",
        "SELECT * FROM users u1 JOIN user_emails ue USING () JOIN users u2 ue USING(first_name, last_name)",
        "SELECT * FROM users u1 JOIN user_emails ue USING() JOIN users u2 ue USING (first_name, last_name)",
    ],
)
def test_join_using_suggests_from_last_table(completer, text):
    position = text.index("()") + 1
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(
        [column("id"), column("email")]
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT * FROM users INNER JOIN orders USING (id,",
        "SELECT * FROM users INNER JOIN orders USING(id,",
    ],
)
def test_join_using_suggests_columns_after_first_column(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [column("id"), column("email")]
    )


@parametrize("completer", completers(casing=False, aliasing=False))
@parametrize(
    "text",
    [
        "SELECT * FROM ",
        "SELECT * FROM users CROSS JOIN ",
        "SELECT * FROM users natural join ",
    ],
)
def test_table_names_after_from(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas_and_from_clause_items()
    )
    assert [c.text for c in result] == [
        "public",
        "orders",
        '"select"',
        "users",
        '"Users"',
        "functions",
        "user_emails",
        "_custom_fun()",
        "custom_fun()",
        "custom_func1()",
        "custom_func2()",
        "set_returning_func(x := , y := )",
    ]


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_auto_escaped_col_names(completer):
    result = get_result(completer, 'SELECT  from "select"', len("SELECT "))
    assert completions_to_set(result) == completions_to_set(
        testdata.columns_functions_and_keywords("select")
    )


@parametrize("completer", completers(aliasing=False))
def test_allow_leading_double_quote_in_last_word(completer):
    result = get_result(completer, 'SELECT * from "sele')

    expected = table('"select"', -5)

    assert expected in result


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT 1::",
        "CREATE TABLE foo (bar ",
        "CREATE FUNCTION foo (bar INT, baz ",
        "ALTER TABLE foo ALTER COLUMN bar TYPE ",
    ],
)
def test_suggest_datatype(text, completer):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas() + testdata.types() + testdata.builtin_datatypes()
    )


@parametrize("completer", completers(casing=False))
def test_suggest_columns_from_escaped_table_alias(completer):
    result = get_result(completer, 'select * from "select" s where s.')
    assert completions_to_set(result) == completions_to_set(testdata.columns("select"))


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_suggest_columns_from_set_returning_function(completer):
    result = get_result(completer, "select  from set_returning_func()", len("select "))
    assert completions_to_set(result) == completions_to_set(
        testdata.columns_functions_and_keywords("set_returning_func", typ="functions")
    )


@parametrize("completer", completers(casing=False))
def test_suggest_columns_from_aliased_set_returning_function(completer):
    result = get_result(
        completer, "select f. from set_returning_func() f", len("select f.")
    )
    assert completions_to_set(result) == completions_to_set(
        testdata.columns("set_returning_func", typ="functions")
    )


@parametrize("completer", completers(casing=False))
def test_join_functions_using_suggests_common_columns(completer):
    text = """SELECT * FROM set_returning_func() f1
              INNER JOIN set_returning_func() f2 USING ("""
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.columns("set_returning_func", typ="functions")
    )


@parametrize("completer", completers(casing=False))
def test_join_functions_on_suggests_columns_and_join_conditions(completer):
    text = """SELECT * FROM set_returning_func() f1
              INNER JOIN set_returning_func() f2 ON f1."""
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [name_join("y = f2.y"), name_join("x = f2.x")]
        + testdata.columns("set_returning_func", typ="functions")
    )


@parametrize("completer", completers())
def test_learn_keywords(completer):
    history = "CREATE VIEW v AS SELECT 1"
    completer.extend_query_history(history)

    # Now that we've used `VIEW` once, it should be suggested ahead of other
    # keywords starting with v.
    text = "create v"
    completions = get_result(completer, text)
    assert completions[0].text == "VIEW"


@parametrize("completer", completers(casing=False, aliasing=False))
def test_learn_table_names(completer):
    history = "SELECT * FROM users; SELECT * FROM orders; SELECT * FROM users"
    completer.extend_query_history(history)

    text = "SELECT * FROM "
    completions = get_result(completer, text)

    # `users` should be higher priority than `orders` (used more often)
    users = table("users")
    orders = table("orders")

    assert completions.index(users) < completions.index(orders)


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_columns_before_keywords(completer):
    text = "SELECT * FROM orders WHERE s"
    completions = get_result(completer, text)

    col = column("status", -1)
    kw = keyword("SELECT", -1)

    assert completions.index(col) < completions.index(kw)


@parametrize("completer", completers(casing=False, qualify=no_qual))
@parametrize(
    "text",
    [
        "SELECT * FROM users",
        "INSERT INTO users SELECT * FROM users u",
        """INSERT INTO users(id, parentid, email, first_name, last_name)
    SELECT *
    FROM users u""",
    ],
)
def test_wildcard_column_expansion(completer, text):
    position = text.find("*") + 1

    completions = get_result(completer, text, position)

    col_list = "id, parentid, email, first_name, last_name"
    expected = [wildcard_expansion(col_list)]

    assert expected == completions


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "SELECT u.* FROM users u",
        "INSERT INTO public.users SELECT u.* FROM users u",
        """INSERT INTO users(id, parentid, email, first_name, last_name)
    SELECT u.*
    FROM users u""",
    ],
)
def test_wildcard_column_expansion_with_alias(completer, text):
    position = text.find("*") + 1

    completions = get_result(completer, text, position)

    col_list = "id, u.parentid, u.email, u.first_name, u.last_name"
    expected = [wildcard_expansion(col_list)]

    assert expected == completions


@parametrize("completer", completers(casing=False))
@parametrize(
    "text,expected",
    [
        (
            "SELECT users.* FROM users",
            "id, users.parentid, users.email, users.first_name, users.last_name",
        ),
        (
            "SELECT Users.* FROM Users",
            "id, Users.parentid, Users.email, Users.first_name, Users.last_name",
        ),
    ],
)
def test_wildcard_column_expansion_with_table_qualifier(completer, text, expected):
    position = len("SELECT users.*")

    completions = get_result(completer, text, position)

    expected = [wildcard_expansion(expected)]

    assert expected == completions


@parametrize("completer", completers(casing=False, qualify=qual))
def test_wildcard_column_expansion_with_two_tables(completer):
    text = 'SELECT * FROM "select" JOIN users u ON true'
    position = len("SELECT *")

    completions = get_result(completer, text, position)

    cols = (
        '"select".id, "select".insert, "select"."ABC", '
        "u.id, u.parentid, u.email, u.first_name, u.last_name"
    )
    expected = [wildcard_expansion(cols)]
    assert completions == expected


@parametrize("completer", completers(casing=False))
def test_wildcard_column_expansion_with_two_tables_and_parent(completer):
    text = 'SELECT "select".* FROM "select" JOIN users u ON true'
    position = len('SELECT "select".*')

    completions = get_result(completer, text, position)

    col_list = 'id, "select".insert, "select"."ABC"'
    expected = [wildcard_expansion(col_list)]

    assert expected == completions


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    ["SELECT U. FROM Users U", "SELECT U. FROM USERS U", "SELECT U. FROM users U"],
)
def test_suggest_columns_from_unquoted_table(completer, text):
    position = len("SELECT U.")
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=False))
def test_suggest_columns_from_quoted_table(completer):
    result = get_result(completer, 'SELECT U. FROM "Users" U', len("SELECT U."))
    assert completions_to_set(result) == completions_to_set(testdata.columns("Users"))


@parametrize("completer", completers(casing=False, aliasing=False))
@parametrize("text", ["SELECT * FROM ", "SELECT * FROM Orders o CROSS JOIN "])
def test_schema_or_visible_table_completion(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas_and_from_clause_items()
    )


@parametrize("completer", completers(casing=False, aliasing=True))
@parametrize("text", ["SELECT * FROM "])
def test_table_aliases(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas() + aliased_rels
    )


@parametrize("completer", completers(casing=False, aliasing=True))
@parametrize("text", ["SELECT * FROM Orders o CROSS JOIN "])
def test_duplicate_table_aliases(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        testdata.schemas()
        + [
            table("orders o2"),
            table("users u"),
            table('"Users" U'),
            table('"select" s'),
            view("user_emails ue"),
            view("functions f"),
            function("_custom_fun() cf"),
            function("custom_fun() cf"),
            function("custom_func1() cf"),
            function("custom_func2() cf"),
            function(
                "set_returning_func(x := , y := ) srf",
                display="set_returning_func(x, y) srf",
            ),
        ]
    )


@parametrize("completer", completers(casing=True, aliasing=True))
@parametrize("text", ["SELECT * FROM Orders o CROSS JOIN "])
def test_duplicate_aliases_with_casing(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [
            schema("PUBLIC"),
            table("Orders O2"),
            table("Users U"),
            table('"Users" U'),
            table('"select" s'),
            view("User_Emails UE"),
            view("Functions F"),
            function("_custom_fun() cf"),
            function("Custom_Fun() CF"),
            function("Custom_Func1() CF"),
            function("custom_func2() cf"),
            function(
                "set_returning_func(x := , y := ) srf",
                display="set_returning_func(x, y) srf",
            ),
        ]
    )


@parametrize("completer", completers(casing=True, aliasing=True))
@parametrize("text", ["SELECT * FROM "])
def test_aliases_with_casing(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [schema("PUBLIC")] + cased_aliased_rels
    )


@parametrize("completer", completers(casing=True, aliasing=False))
@parametrize("text", ["SELECT * FROM "])
def test_table_casing(completer, text):
    result = get_result(completer, text)
    assert completions_to_set(result) == completions_to_set(
        [schema("PUBLIC")] + cased_rels
    )


@parametrize("completer", completers(casing=False))
@parametrize(
    "text",
    [
        "INSERT INTO users ()",
        "INSERT INTO users()",
        "INSERT INTO users () SELECT * FROM orders;",
        "INSERT INTO users() SELECT * FROM users u cross join orders o",
    ],
)
def test_insert(completer, text):
    position = text.find("(") + 1
    result = get_result(completer, text, position)
    assert completions_to_set(result) == completions_to_set(testdata.columns("users"))


@parametrize("completer", completers(casing=False, aliasing=False))
def test_suggest_cte_names(completer):
    text = """
        WITH cte1 AS (SELECT a, b, c FROM foo),
             cte2 AS (SELECT d, e, f FROM bar)
        SELECT * FROM
    """
    result = get_result(completer, text)
    expected = completions_to_set(
        [
            Completion("cte1", 0, display_meta="table"),
            Completion("cte2", 0, display_meta="table"),
        ]
    )
    assert expected <= completions_to_set(result)


@parametrize("completer", completers(casing=False, qualify=no_qual))
def test_suggest_columns_from_cte(completer):
    result = get_result(
        completer,
        "WITH cte AS (SELECT foo, bar FROM baz) SELECT  FROM cte",
        len("WITH cte AS (SELECT foo, bar FROM baz) SELECT "),
    )
    expected = [
        Completion("foo", 0, display_meta="column"),
        Completion("bar", 0, display_meta="column"),
    ] + testdata.functions_and_keywords()

    assert completions_to_set(expected) == completions_to_set(result)


@parametrize("completer", completers(casing=False, qualify=no_qual))
@parametrize(
    "text",
    [
        "WITH cte AS (SELECT foo FROM bar) SELECT * FROM cte WHERE cte.",
        "WITH cte AS (SELECT foo FROM bar) SELECT * FROM cte c WHERE c.",
    ],
)
def test_cte_qualified_columns(completer, text):
    result = get_result(completer, text)
    expected = [Completion("foo", 0, display_meta="column")]
    assert completions_to_set(expected) == completions_to_set(result)


@parametrize(
    "keyword_casing,expected,texts",
    [
        ("upper", "SELECT", ("", "s", "S", "Sel")),
        ("lower", "select", ("", "s", "S", "Sel")),
        ("auto", "SELECT", ("", "S", "SEL", "seL")),
        ("auto", "select", ("s", "sel", "SEl")),
    ],
)
def test_keyword_casing_upper(keyword_casing, expected, texts):
    for text in texts:
        completer = testdata.get_completer({"keyword_casing": keyword_casing})
        completions = get_result(completer, text)
        assert expected in [cpl.text for cpl in completions]


@parametrize("completer", completers())
def test_keyword_after_alter(completer):
    text = "ALTER TABLE users ALTER "
    expected = Completion("COLUMN", start_position=0, display_meta="keyword")
    completions = get_result(completer, text)
    assert expected in completions


@parametrize("completer", completers())
def test_set_schema(completer):
    text = "SET SCHEMA "
    result = get_result(completer, text)
    expected = completions_to_set([schema("'public'")])
    assert completions_to_set(result) == expected


@parametrize("completer", completers())
def test_special_name_completion(completer):
    result = get_result(completer, "\\t")
    assert completions_to_set(result) == completions_to_set(
        [
            Completion(
                text="\\timing",
                start_position=-2,
                display_meta="Toggle timing of commands.",
            )
        ]
    )
